#!/usr/bin/env python3
import logging
import os
import socket
import struct
try:
    import fcntl
except ImportError:
    fcntl = None
from xml.etree import ElementTree

import dbus
from _dbus_bindings import BUS_DAEMON_NAME, BUS_DAEMON_PATH, BUS_DAEMON_IFACE

from .errors import NoBluezDbus, NoAdapter, UnpoweredAdapter

log = logging.getLogger(__name__)

def get_bt_addr_for_interface(iface):
    if not fcntl:
        raise NoAdapter("fcntl not available on this platform")

    # Create a Bluetooth socket
    sock = socket.socket(socket.AF_BLUETOOTH, socket.SOCK_RAW, socket.BTPROTO_HCI)

    # Get the Bluetooth address
    # SIOCGIFHWADDR is 0x8927
    # struct.pack formats: '256s' means 256-char string
    try:
        address = fcntl.ioctl(sock.fileno(), 0x8927, struct.pack("256s", iface.encode()[:15]))
    except OSError as e:
        if e.errno == 19: # No such device
            raise NoAdapter from e
        else:
            log.exception("cannot ioctl bt")
            raise
    finally:
        # Close the socket
        sock.close()

    # Extract the Bluetooth address from the returned bytes
    bt_address = ':'.join(['%02X' % b for b in address[18:24]])

    log.debug("BT: addr of %s is %s", iface, bt_address)
    return bt_address


def get_bluetooth_addresses():
    addresses = {}

    # Check for Bluetooth interfaces
    bluetooth_dir = '/sys/class/bluetooth'
    if not os.path.exists(bluetooth_dir):
        raise NoAdapter

    # Iterate through all Bluetooth interfaces
    ifaces = os.listdir(bluetooth_dir)
    log.info("BT interfaces: %s", ifaces)
    for interface in ifaces:
        try:
            bt_address = get_bt_addr_for_interface(interface)
            addresses[interface] = bt_address
        except Exception as e:
            addresses[interface] = f"Error: {str(e)}"

    return addresses

def get_local_bt_address():
    """Check if there is a powered on Bluetooth device and return its address.
       This is a blocking method"""

    #addrs = get_bluetooth_addresses()
    #log.info("Local BT addresses: %s", addrs)
    #return list(addrs.values())[0]

    available = False
    bus_name = "org.bluez"
    timeout = 2  # 2 seconds seems to be enough to start a bus service
    bus = dbus.SystemBus()

    try:
        _start_bus(bus_name, timeout)
    except dbus.exceptions.DBusException as e:
        raise NoBluezDbus(e)
    else:
        available_bt = _get_available_bt()
        for bt in available_bt:
            adapter = dbus.Interface(bus.get_object("org.bluez", bt), "org.freedesktop.DBus.Properties")
            power = adapter.Get("org.bluez.Adapter1", "Powered")
            if power:
                available = adapter.Get("org.bluez.Adapter1", "Address")
                break

        if len(available_bt) == 0:
            # Not a single BT adapter available in the system
            raise NoAdapter

        elif not available:
            # Every BT adapters are powered off
            raise UnpoweredAdapter

        return available


def _start_bus(bus_name, timeout, flags=0):
    """Manually start the bus, so we can set a custom timeout"""
    bus = dbus.SystemBus()
    bus.call_blocking(BUS_DAEMON_NAME, BUS_DAEMON_PATH,
                      BUS_DAEMON_IFACE,
                      'StartServiceByName',
                      'su', (bus_name, flags), timeout=timeout)


def _get_available_bt():
    """Returns the list of available Bluetooth"""
    available_bt = []
    bus_name = "org.bluez"
    object_path = "/org/bluez"
    bus = dbus.SystemBus()
    obj = bus.get_object(bus_name, object_path)
    iface = dbus.Interface(obj, 'org.freedesktop.DBus.Introspectable')
    xml_string = iface.Introspect()
    for child in ElementTree.fromstring(xml_string):
        if child.tag == 'node':
            bt = '/'.join((object_path, child.attrib['name']))
            available_bt.append(bt)
    log.info("Available Bluetooth interfaces: %s", available_bt)
    return available_bt



#!/usr/bin/env python
#    Copyright 2014 Tobias Mueller <muelli@cryptobitch.de>
#    Copyright 2014 Andrei Macavei <andrei.macavei89@gmail.com>
#    Copyright 2015 Jody Hansen <jobediah.hansen@gmail.com>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.
import logging
import os

import dbus
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GObject

if __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    __package__ = str('keysign')

from .AvahiConstants import AvahiConstants as avahi
DBusGMainLoop(set_as_default=True)


class AvahiPublisher:

    def __init__(self,
            service_name='Demo Service',
            service_type='_demo._tcp',
            service_port=8899,
            service_txt={},
            domain='',
            host=''):
        self.log = logging.getLogger(__name__)
        #self.loop = loop or DBusGMainLoop()
        self.bus = dbus.SystemBus()
        self.server = dbus.Interface(
            self.bus.get_object( avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER ),
            avahi.DBUS_INTERFACE_SERVER )

        self.service_name = service_name
        #See http://www.dns-sd.org/ServiceTypes.html
        self.service_type = service_type
        self.service_port = service_port
        self.service_txt = avahi.dict_to_txt_array(service_txt) #TXT record for the service
        self.domain = domain # Domain to publish on, default to .local
        self.host = host # Host to publish records for, default to localhost

        self.group = None
        # Counter so we only rename after collisions a sensible number of times
        self.rename_count = 12


    def add_service(self):
        if self.group is None:
            group = dbus.Interface(
                    self.bus.get_object(
                        avahi.DBUS_NAME, self.server.EntryGroupNew()),
                    avahi.DBUS_INTERFACE_ENTRY_GROUP)
            group.connect_to_signal('StateChanged',
                self.entry_group_state_changed)

            self.group = group

        self.log.info("Adding service '%s' of type '%s' with fpr '%s'",
            self.service_name, self.service_type, self.service_txt)

        group = self.group
        group.AddService(
                avahi.IF_UNSPEC,    #interface
                avahi.PROTO_UNSPEC, #protocol
                dbus.UInt32 (0),    #flags
                self.service_name, self.service_type,
                self.domain, self.host,
                dbus.UInt16 (self.service_port),
                self.service_txt)
        group.Commit()

    def remove_service(self):
        '''Publishes services to be removed with name, stype, and domain.'''
        self.log.info("Removing with fpr '%s'", self.service_txt)
        if not self.group is None:
            self.group.Reset()

    def server_state_changed(self, state):
        if state == avahi.SERVER_COLLISION:
            self.log.warn("Server name collision (%s)", self.service_name)
            self.remove_service()
        elif state == avahi.SERVER_RUNNING:
            self.add_service()

    def entry_group_state_changed(self, state, error):
        self.log.debug("state change: %i", state)

        if state == avahi.ENTRY_GROUP_ESTABLISHED:
            self.log.info("Service established.")

        elif state == avahi.ENTRY_GROUP_COLLISION:
            self.rename_count -= 1
            if self.rename_count > 0:
                name = self.server.GetAlternativeServiceName(self.service_name)
                self.log.warn("Service name collision, changing name to '%s'",
                    name)
                self.remove_service()
                self.add_service()

            else:
                # FIXME: max_renames is not defined. We probably want to restructure
                # this a little bit, anyway. i.e. have a self.max_renames
                # and a self.rename_count or so
                m = "No suitable service name found after %i retries, exiting."
                self.log.error(m, self.max_renames)
                raise RuntimeError(m % self.max_renames)

        elif state == avahi.ENTRY_GROUP_FAILURE:
            m = "Error in group state changed %s"
            self.log.error(m, error)
            raise RuntimeError(m % error)


DBusGMainLoop( set_as_default=True )
if __name__ == '__main__':
    ap = AvahiPublisher()
    ap.add_service()

    main_loop = GObject.MainLoop()
    bus = dbus.SystemBus()

    server = dbus.Interface(
            bus.get_object( avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER ),
            avahi.DBUS_INTERFACE_SERVER )

    server.connect_to_signal( "StateChanged", ap.server_state_changed )
    ap.server_state_changed( server.GetState() )


    try:
        main_loop.run()
    except KeyboardInterrupt:
        pass

    if not ap.group is None:
        ap.group.Free()




#!/usr/bin/env python
#    Copyright 2014 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

"""This is a very simple QR Code generator which scans your GnuPG keyring
for keys and selects the one matching your input
"""
import logging
import os
import sys

import gi
gi.require_version("Gtk", "4.0")
from gi.repository import Gtk

if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')

from .gpgmeh import get_usable_keys

if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')


from .QRCode import QRImage


def main():
    import sys
    key = sys.argv[1]
    # Heh, we take the first key here. Maybe we should raise a warning
    # or so, when there is more than one key.
    key = list(get_usable_keys(pattern=key))[0]
    fpr = key.fingerprint
    data = 'OPENPGP4FPR:' + fpr
    
    w = Gtk.Window()
    w.connect("delete-event", Gtk.main_quit)
    w.set_default_size(100,100)
    v = Gtk.VBox()
    label = Gtk.Label(data)
    qr = QRImage(data)
    v.add(label)
    v.add(qr)
    w.add(v)
    w.show_all()
    Gtk.main()

if __name__ == '__main__':
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
                        format='%(name)s (%(levelname)s): %(message)s')
    main()



#!/usr/bin/env python
import logging
import os

import gi
gi.require_version('Gtk', '4.0')
#gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from gi.repository import GObject  # for __gsignals__
from gi.repository import GLib  # for markup_escape_text

if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')

from .gpgmeh import get_usable_keys
from .i18n import _
from .util import fix_infobar

log = logging.getLogger(__name__)

class ListBoxRowWithKey(Gtk.ListBoxRow):
    "A simple extension of a Gtk.ListBoxRow to also hold a key object"

    def __init__(self, key):
        super(ListBoxRowWithKey, self).__init__()
        self.key = key

        s = self.format(key)
        label = Gtk.Label(label=s, use_markup=True, xalign=0)
        self.set_child(label)

    @staticmethod
    def glib_markup_escape_text_to_text(s):
        """A helper function to return the text type
        markup_escape_text returns a "str" which is
        a binary type in python2.
        This function tries to decode the returned
        str object.  It will fail in Python3.
        """
        m = GLib.markup_escape_text(s)
        try:
            ret = m.decode('utf-8')
        except AttributeError:
            # We are in Python3 land. All is fine.
            ret = m
        return ret
        

    @classmethod
    def format_uid(cls, uid):
        "Returns a pango string for a gpgmeh.UID"
        fmt = "{name}\t<i>{email}</i>\t<small>{expiry}</small>"

        items = ('name', 'email', 'expiry')
        format_dict = {k: ""+(uid._asdict()[k] or "")
                          for k in items}
        log.info("format dicT: %r", format_dict)
        d = {k: (log.debug("handling kv: %r %r", k, v),
                  cls.glib_markup_escape_text_to_text(
                    "{}".format(v)))[1]
             for k, v in format_dict.items()}
        log.info("Formatting UID %r", d)
        s = fmt.format(**d)
        log.info("Formatted UID: %r", s)
        return s


    @classmethod
    def format(cls, key):
        "Returns a pango string for a gpgmeh.Key"
        fmt  = "{created} "
        fmt  = "<b>{fingerprint}</b>\n"
        fmt += "\n".join((cls.format_uid(uid) for uid in key.uidslist))
        fmt += "\n<small>" + _("Expires: ") + " {expiry}</small>"

        d = {k: GLib.markup_escape_text("{}".format(v))
             for k,v in key._asdict().items()}
        log.info("Formatting key %r", d)
        s = fmt.format(**d)
        log.info("Formatted key: %r", s)
        return s


class KeyListWidget(Gtk.Box):
    """A Gtk Widget representing a list of OpenPGP Keys
    
    It shows the keys you provide in a ListBox and emits a
    `key-activated` or `key-selected` signal when the user
    "activated" or "selected" a key. "Activating" is Gtk speak for
    double-clicking (or pressing space, enter, ...) on an entry.
    It is also possible that the widget emits that signal on a single
    click if so configured.  "Selected" means that the user switched
    to an entry, e.g. by clicking it or pressing up or down.

    If you don't provide any keys, the widget will not behave nicely
    and potentially display a user facing warning. Or not.
    """
    __gsignals__ = {
        str('key-activated'): (GObject.SignalFlags.RUN_LAST, None,
                               # (ListBoxRowWithKey.__gtype__,)
                               (object,)),
                               # The activated key
        str('key-selected'): (GObject.SignalFlags.RUN_LAST, None,
                               # (ListBoxRowWithKey.__gtype__,)
                               (object,)),
                               # The selected key
    }

    def __init__(self, keys, builder=None):
        "Sets the widget up with the given keys"
        super(KeyListWidget, self).__init__()
        self.log = logging.getLogger(__name__)
        self.log.debug("KLW with keys: %r", keys)

        thisdir = os.path.dirname(os.path.abspath(__file__))
        widget_name = 'keylistbox'
        if not builder:
            builder = Gtk.Builder()
            builder.add_objects_from_file(
                os.path.join(thisdir, 'send4.ui'),
                [widget_name])
        widget = builder.get_object(widget_name)
        old_parent = widget.get_parent()
        if old_parent:
            old_parent.remove(widget)
        self.append(widget)

        self.listbox = builder.get_object("keys_listbox")
        self.code_spinner = builder.get_object("code_spinner")
        self.ib_internet = builder.get_object('infobar_internet')
        fix_infobar(self.ib_internet)
        self.label_ib_internet = builder.get_object('label_internet')

        self.ib_import_okay = builder.get_object('infobar_import_okay')
        self.ib_import_error = builder.get_object('infobar_import_error')
        self.ib_import_error_no_new_sigs = builder.get_object('infobar_import_error_no_new_signatures')
        assert self.ib_import_okay
        assert self.ib_import_error
        assert self.ib_import_error_no_new_sigs
        fix_infobar(self.ib_import_okay)
        fix_infobar(self.ib_import_error)
        self.label_ib_import_okay = builder.get_object('label_import_okay')
        self.image_ib_import_okay = builder.get_object('image_import_okay')
        self.label_ib_import_error = builder.get_object('label_import_error')
        self.image_ib_import_error = builder.get_object('image_import_error')
        self.button_ib_import_error = builder.get_object('import_error_details_button')
        self.button_ib_import_okay = builder.get_object('return_signature')

        if len(list(keys)) <= 0:
            infobar = builder.get_object("infobar")
            infobar.show()
            l = Gtk.Label("You don't have any OpenPGP keys")
            self.listbox.append(l)
        else:
            for key in keys:
                self.log.debug("Adding key: %r", key)
                lbr = ListBoxRowWithKey(key)
                lbr.props.margin_bottom = 5
                self.listbox.append(lbr)
            self.listbox.connect('row-activated', self.on_row_activated)
            self.listbox.connect('row-selected', self.on_row_selected)


    def on_row_activated(self, keylistwidget, row):
        if row:
            self.emit('key-activated', row.key)

    def on_row_selected(self, keylistwidget, row):
        if row:
            self.emit('key-selected', row.key)


class App(Gtk.Application):
    def __init__(self, *args, **kwargs):
        super(App, self).__init__(*args, **kwargs)
        self.connect('activate', self.on_activate)
        self.kpw = None

    def on_activate(self, app):
        window = Gtk.ApplicationWindow(application=app)
        window.set_title("Key List")

        if not self.kpw:
            self.kpw = KeyListWidget(get_usable_keys())
        self.kpw.connect('key-activated', self.on_key_activated)
        self.kpw.connect('key-selected', self.on_key_selected)
        window.set_child(self.kpw)

        window.present()
        self.add_window(window)

    def on_key_activated(self, keylistwidget, row):
        self.get_windows()[0].get_display().beep()
        print ("Row activated! %r" % (row,))

    def on_key_selected(self, keylistwidget, row):
        print ("Row selected! %r" % (row,))

    def run(self, args):
        if not args:
            args = [""]
        keys = list(get_usable_keys(pattern=args[0]))
        self.kpw = KeyListWidget(keys)
        super(App, self).run()

if __name__ == "__main__":
    import sys
    logging.basicConfig(level=logging.DEBUG)
    app = App()
    app.run(sys.argv[1:])



#!/usr/bin/env python
# encoding: utf-8
#    Copyright 2016 Andrei Macavei <andrei.macavei89@gmail.com>
#    Copyright 2017 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

from datetime import date, datetime
import signal
import sys
import argparse
import logging
import os

import gi
gi.require_version('Gtk', '4.0')

from gi.repository import Gtk, GLib, GdkPixbuf
from gi.repository import GObject


if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')


from .gpgmeh import get_usable_keys
from .scan_barcode import ScalingImage
from .util import format_fingerprint

log = logging.getLogger(__name__)


#FIXME: remove the temporary keyword args after updating Key class
#with length and creation_time fields
def format_key_header(fpr, length='2048', creation_time=None):
    if creation_time == None:
        creation_time = datetime.strptime('01011970', "%d%m%Y").date()
    try:
        creation = date.fromtimestamp(float(creation_time))
    except TypeError as e:
        # This might be the case when the creation_time is already a timedate
        creation = creation_time

    key_header = format_fingerprint(fpr).replace('\n', '  ')
    return key_header

def format_uidslist(uidslist):
    result = ""
    for uid in uidslist:
        uidstr = GLib.markup_escape_text(uid.uid)
        result += ("{}\n".format(uidstr))

    return result



class PreSignWidget(Gtk.Box):
    """A widget for obtaining a key fingerprint.

    The fingerprint can be obtain by inserting it into
    a text entry, or by scanning a barcode with the
    built-in camera.
    """

    __gsignals__ = {
        str('sign-key-confirmed'): (GObject.SignalFlags.RUN_LAST, None,
                                    (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self, key, pixbuf=None, builder=None):
        super(PreSignWidget, self).__init__()
        thisdir = os.path.dirname(os.path.abspath(__file__))
        widget_name = 'keyconfirmbox'
        if not builder:
            builder = Gtk.Builder()
            builder.add_objects_from_file(
                os.path.join(thisdir, 'receive4.ui'),
                [widget_name, 'confirm-button-image', 'error_dialog'])
        widget = builder.get_object(widget_name)
        parent = widget.get_parent()
        if parent:
            parent.remove(widget)
        self.append(widget)

        confirm_btn = builder.get_object("confirm_sign_button")
        confirm_btn.connect("clicked", self.on_confirm_button_clicked)

        self.key = key

        keyIdsLabel = builder.get_object("key_ids_label")
        # Weird. The glade file defines can_focus = False, but it's set to True...
        keyIdsLabel.set_focusable(False)
        keyIdsLabel.set_markup(format_key_header(self.key.fingerprint))

        uidsLabel = builder.get_object("uids_label")
        # FIXME: Check why Builder thinks the widget can focus when the glade file says no
        uidsLabel.set_focusable(False)
        markup = format_uidslist(self.key.uidslist)
        uidsLabel.set_markup(markup)

        imagebox = builder.get_object("imagebox")
        child = imagebox.get_first_child()
        while child:
            next_child = child.get_next_sibling()
            imagebox.remove(child)
            child = next_child
        imagebox.append(ScalingImage(pixbuf=pixbuf))

        # We save the reference here to expose this infobar to the caller.
        # This is a bit ugly, because it makes this implementation detail part of the
        # API. The infobar should probably be part of the caller's responsibility,
        # i.e. not part of this widget.
        self.infobar_success = builder.get_object('infobar_certifications_produced')
        self.infobar_errors = builder.get_object('infobar_certifications_errors')
        self.infobar_save_as_button = builder.get_object('btn_local_import_save_as')
        self.infobar_import_button = builder.get_object('btn_local_import')
        self.infobar_show_error_button = builder.get_object('btn_show_error_details')

        ib_error_show = self.infobar_errors.show
        
        self.error_dialog = builder.get_object('error_dialog')
        self.error_dialog_base_text = self.error_dialog.get_property('secondary-text')
        self.error_dialog.connect("response", lambda d, r: d.hide())

        def show_error(btn):
            if hasattr(self.infobar_errors, 'exception'):
                exception = self.infobar_errors.exception
                self.error_dialog.set_transient_for(self.get_root())
                self.error_dialog.set_property('secondary-text',
                    self.error_dialog_base_text % (str(exception),))
                self.error_dialog.present()

        self.infobar_show_error_button.connect("clicked", show_error)

        def show(exception):
            self.infobar_errors.exception = exception
            ib_error_show()
            
        self.infobar_errors.show = show

    def on_confirm_button_clicked(self, buttonObject, *args):
        self.emit('sign-key-confirmed', self.key, *args)



class PreSignApp(Gtk.Application):
    def __init__(self, *args, **kwargs):
        super(PreSignApp, self).__init__(*args, **kwargs)
        self.connect('activate', self.on_activate)
        self.psw = None

        self.log = logging.getLogger(__name__)

    def on_activate(self, app):
        window = Gtk.ApplicationWindow(application=app)
        window.set_title("Key Pre Sign Widget")
        # window.set_size_request(600, 400)

        if not self.psw:
            self.psw = PreSignWidget()

        self.psw.connect('sign-key-confirmed', self.on_sign_key_confirmed)
        window.set_child(self.psw)

        window.present()
        self.add_window(window)

    def on_sign_key_confirmed(self, keyPreSignWidget, *args):
        self.log.debug ("Sign key confirmed!")

    def run(self, args):
        if not args:
            args = [""]
        key = get_usable_keys (pattern=args[0])[0]
        if len(args) >= 2:
            image_fname = args[1]
            log.debug("Trying to load pixbuf from %r", image_fname)
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(image_fname)
        else:
            pixbuf = None
        self.psw = PreSignWidget(key, pixbuf=pixbuf)
        super(PreSignApp, self).run()


if __name__ == "__main__":
    import sys
    logging.basicConfig(level=logging.DEBUG)
    app = PreSignApp()
    app.run(sys.argv[1:])



#!/usr/bin/env python
# (c) 2003 Gustavo J A M Carneiro gjc at inescporto.pt
#     2004-2005 Filip Van Raemdonck
#
# http://www.daa.com.au/pipermail/pygtk/2003-August/005775.html
# Message-ID: <1062087716.1196.5.camel@emperor.homelinux.net>
#     "The license is whatever you want."
#
# This file was downloaded from http://www.sysfs.be/downloads/
# Adaptions 2009-2010 by Martin Renold:
# - let KeyboardInterrupt through
# - print traceback to stderr before showing the dialog
# - nonzero exit code when hitting the "quit" button
# - suppress more dialogs while one is already active
# - fix Details button when a context in the traceback is None
# - remove email features
# - fix lockup with dialog.run(), return to mainloop instead
# see also http://faq.pygtk.org/index.py?req=show&file=faq20.010.htp
# (The license is still whatever you want.)

import inspect
import linecache
import pydoc
import sys
import traceback
from io import StringIO as StringIO
from gettext import gettext as _
import os

try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus
import textwrap

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Pango

#with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
#                       '_version.py')) as f:
#    # This should define __version__
#    exec(f.read())
#VERSION = __version__
# We are ignoring the actual version for now, because it's only used
# to determine whether it's a -dev version to then show a "report"
# button. Let's show this button unconditionally for now.
# Once it's too annoying, we get rid of it.
VERSION = "0.0.0.1-dev"

# Function that will be called when the user presses "Quit"
# Return True to confirm quit, False to cancel
quit_confirmation_func = None

RESPONSE_QUIT = 1
RESPONSE_SEARCH = 2
RESPONSE_REPORT = 3

def analyse_simple(exctyp, value, tb):
    trace = StringIO()
    traceback.print_exception(exctyp, value, tb, None, trace)
    return trace


def lookup(name, frame, lcls):
    '''Find the value for a given name in the given frame'''
    if name in lcls:
        return 'local', lcls[name]
    elif name in frame.f_globals:
        return 'global', frame.f_globals[name]
    elif '__builtins__' in frame.f_globals:
        builtins = frame.f_globals['__builtins__']
        if type(builtins) is dict:
            if name in builtins:
                return 'builtin', builtins[name]
        else:
            if hasattr(builtins, name):
                return 'builtin', getattr(builtins, name)
    return None, []


def analyse(exctyp, value, tb):
    import tokenize
    import keyword
    import platform
    #import application

    #app = application.get_app()

    trace = StringIO()
    nlines = 3
    frecs = inspect.getinnerframes(tb, nlines)

    #trace.write('GNOME Keysign version: %s\n' % app.version)
    trace.write('System information: %s\n' % platform.platform())
    trace.write('Python information: %s\n' % sys.version)
    #trace.write('Using: %s\n' % (get_libs_version_string(),))

    trace.write('Traceback (most recent call last):\n')
    for frame, fname, lineno, funcname, context, cindex in frecs:
        trace.write('  File "%s", line %d, ' % (fname, lineno))
        args, varargs, varkw, lcls = inspect.getargvalues(frame)

        def readline(lno=[lineno], *args):
            if args:
                print(args)

            try:
                return linecache.getline(fname, lno[0])
            finally:
                lno[0] += 1
        all, prev, name, scope = {}, None, '', None
        for ttype, tstr, stup, etup, line in tokenize.generate_tokens(readline):
            if ttype == tokenize.NAME and tstr not in keyword.kwlist:
                if name:
                    if name[-1] == '.':
                        try:
                            val = getattr(prev, tstr)
                        except AttributeError:
                            # XXX skip the rest of this identifier only
                            break
                        name += tstr
                else:
                    assert not name and not scope
                    scope, val = lookup(tstr, frame, lcls)
                    name = tstr
                if val is not None:
                    prev = val
            elif tstr == '.':
                if prev:
                    name += '.'
            else:
                if name:
                    all[name] = (scope, prev)
                prev, name, scope = None, '', None
                if ttype == tokenize.NEWLINE:
                    break

        try:
            details = inspect.formatargvalues(args, varargs, varkw, lcls, formatvalue=lambda v: '=' + pydoc.text.repr(v))
        except:
            # seen that one on Windows (actual exception was KeyError: self)
            details = '(no details)'
        trace.write(funcname + details + '\n')
        if context is None:
            context = ['<source context missing>\n']
        trace.write(''.join(['    ' + x.replace('\t', '  ') for x in filter(lambda a: a.strip(), context)]))
        if len(all):
            trace.write('  variables: %s\n' % str(all))

    trace.write('%s: %s' % (exctyp.__name__, value))
    return trace


def _info(exctyp, value, tb):
    global exception_dialog_active
    if exctyp is KeyboardInterrupt:
        return original_excepthook(exctyp, value, tb)
    sys.stderr.write(analyse_simple(exctyp, value, tb).getvalue())
    if exception_dialog_active:
        return

    exception_dialog_active = True
    # Create the dialog
    dialog = Gtk.MessageDialog(message_type=Gtk.MessageType.WARNING)
    dialog.set_title(_("Bug Detected"))

    primary = _(
        "<big><b>A programming error has been detected.</b></big>"
    )
    secondary = _(
        "You may be able to ignore this error and carry on working, "
        "but you may get unexpected results.\n\n"
        "Please tell the developers about this using the issue tracker "
        "if no-one else has reported it yet."
    )
    dialog.set_markup(primary)
    dialog.set_property('secondary-text', secondary)

    dialog.add_button(_("Search Tracker..."), RESPONSE_SEARCH)
    if "-" in VERSION:  # only development and prereleases
        dialog.add_button(_("Report..."), RESPONSE_REPORT)
        dialog.set_response_sensitive(RESPONSE_REPORT, False)
    dialog.add_button(_("Ignore Error"), Gtk.ResponseType.CLOSE)
    dialog.add_button(_("Quit GNOME Keysign"), RESPONSE_QUIT)

    # Add an expander with details of the problem to the dialog
    def expander_cb(expander, *ignore):
        # Ensures that on deactivating the expander, the dialog is resized down
        if expander.get_expanded():
            dialog.set_resizable(True)
        else:
            dialog.set_resizable(False)
    details_expander = Gtk.Expander()
    details_expander.set_label(_("Details..."))
    details_expander.connect("notify::expanded", expander_cb)

    textview = Gtk.TextView()
    textview.set_editable(False)

    sw = Gtk.ScrolledWindow()
    sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
    sw.set_child(textview)

    # Set window sizing so that it's always at least 600 pixels wide, and
    # increases by 300 pixels in height once the details panel is open
    sw.set_size_request(0, 300)
    dialog.set_size_request(600, 0)

    details_expander.set_child(sw)
    dialog.get_content_area().append(details_expander)

    # Get the traceback and set contents of the details
    try:
        trace = analyse(exctyp, value, tb).getvalue()
    except:
        try:
            trace = _("Exception while analyzing the exception.") + "\n"
            trace += analyse_simple(exctyp, value, tb).getvalue()
        except:
            trace = _("Exception while analyzing the exception.")
    buf = textview.get_buffer()
    trace = "\n".join(["```python", trace, "```"])
    buf.set_text(trace)
    ## Would be nice to scroll to the bottom automatically, but @#&%*@
    #first, last = buf.get_bounds()
    #buf.place_cursor(last)
    #mark = buf.get_insert()
    ##buf.scroll_mark_onscreen()
    ##textview.scroll_mark_onscreen(buf.get_insert(), 0)
    #textview.scroll_to_mark(mark, 0.0)

    # Connect callback and present the dialog
    dialog.connect('response', _dialog_response_cb, trace, exctyp, value)
    #dialog.set_modal(True) # this might actually be contra-productive...
    dialog.present()
    # calling dialog.run() here locks everything up in some cases, so
    # we just return to the main loop instead


def _dialog_response_cb(dialog, resp, trace, exctyp, value):
    global exception_dialog_active

    if resp == RESPONSE_QUIT:
        if not quit_confirmation_func:
            sys.exit(1)  # Exit code is important for IDEs
        else:
            if quit_confirmation_func():
                sys.exit(1)  # Exit code is important for IDEs
            else:
                dialog.destroy()
                exception_dialog_active = False
    elif resp == RESPONSE_SEARCH:
        search_url = (
            "https://github.com/GNOME-Keysign/gnome-keysign/search"
            "?utf8=%E2%9C%93"
            "&q={}+{}"
            "&type=Issues"
        ).format(
            quote_plus(exctyp.__name__, "/"),
            quote_plus(str(value), "/")
        )
        Gtk.show_uri(None, search_url, None)
        if "-" in VERSION:
            dialog.set_response_sensitive(RESPONSE_REPORT, True)
    elif resp == RESPONSE_REPORT:
        #TRANSLATORS: Crash report template for github, preceding a traceback.
        #TRANSLATORS: Please ask users kindly to supply at least an English
        #TRANSLATORS: title if they are able.
        body = _(u"""\
            #### Description

            Give this report a short descriptive title.
            Use something like
            "{feature-that-broke}: {what-went-wrong}"
            for the title, if you can.
            Then please replace this text
            with a longer description of the bug.
            Screenshots or videos are great, too!

            #### Steps to reproduce

            Please tell us what you were doing
            when the error message popped up.
            If you can provide step-by-step instructions
            on how to reproduce the bug,
            that's even better.

            #### Traceback
        """)
        body = "\n\n".join([
            "".join(textwrap.wrap(p, sys.maxsize))
            for p in textwrap.dedent(body).split("\n\n")
        ] + [trace])
        report_url = (
            "https://github.com/GNOME-Keysign/gnome-keysign/issues/new"
            "?title={title}"
            "&body={body}"
        ).format(
            title="",
            body=quote_plus(body.encode("utf-8"), "/"),
        )
        Gtk.show_uri(None, report_url, None)
    else:
        dialog.destroy()
        exception_dialog_active = False


original_excepthook = sys.excepthook
sys.excepthook = _info
exception_dialog_active = False


if __name__ == '__main__':
    import sys
    import os

    def _test_button_clicked_cb(*a):
        class _TestException (Exception):
            pass
        raise _TestException("That was supposed to happen.")

    win = Gtk.Window()
    win.set_size_request(200, 150)
    win.set_title(os.path.basename(sys.argv[0]))
    btn = Gtk.Button(label="Break it")
    btn.connect("clicked", _test_button_clicked_cb)
    win.set_child(btn)
    win.connect("destroy", lambda *a: Gtk.main_quit())
    win.present()
    Gtk.main()



#!/usr/bin/env python
#    Copyright 2017 Ludovico de Nittis <aasonykk+gnome@gmail.com>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

from binascii import hexlify
from textwrap import dedent
import logging
import os
from builtins import input

from wormhole.cli.public_relay import RENDEZVOUS_RELAY
from wormhole.errors import TransferError, ServerConnectionError, WrongPasswordError, LonelyError
import wormhole

if __name__ == "__main__":
    from twisted.internet import gireactor
    gireactor.install()
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks

if __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    __package__ = str('keysign')

from .gpgmeh import get_usable_keys, get_public_key_data
from .util import encode_message, decode_message

log = logging.getLogger(__name__)


class WormholeOffer:
    def __init__(self, key, app_id=None):
        self.message_def = None
        self.key = key
        if not app_id:
            # the following id is needed for interoperability with wormhole cli
            app_id = "lothar.com/wormhole/text-or-file-xfer"
        self.w = wormhole.create(app_id, RENDEZVOUS_RELAY, reactor)

    @inlineCallbacks
    def allocate_code(self, code=None):
        if code:
            self.w.set_code(code)
        else:
            # ServerConnectionError may be raised
            self.w.allocate_code()
            code = yield self.w.get_code()
        log.info("Invitation Code: %s", code)
        wormhole_data = "WORM={0}".format(code)
        return code, wormhole_data

    @inlineCallbacks
    def start(self):
        log.info("Wormhole: Sending a message")
        try:
            verifier = yield self.w.get_verifier()
            # TODO maybe we can show it to the user and ask for a confirm that is the right one
            ver_ascii = hexlify(verifier).decode("ascii")
            log.info("Verified key: %s", ver_ascii)

            key_data = get_public_key_data(self.key.fingerprint)
            kd_decoded = key_data.decode('utf-8')
            # The message needs to be encoded as a json with "message" and "offer" for ensures
            # wormhole cli interoperability
            offer = {"message": kd_decoded}
            data = {"offer": offer}
            m = encode_message(data)
            self.w.send_message(m)

            # wait for reply
            # TODO add a timeout?
            msg = yield self.w.get_message()

            log.info("Got data, %d bytes" % len(msg))
            success, error_msg = self._check_received(msg)
            self.stop()
            ret_val = success, error_msg

        except (ServerConnectionError, WrongPasswordError) as e:
            error = dedent(e.__doc__)
            log.error("Error: %s" % error)
            success = False
            ret_val = success, e
        except LonelyError as le:
            log.info("Lonely, close() was called before the peer connection could be established")
            success = False
            ret_val = success, le
        except Exception as e:
            error = dedent(e.__doc__)
            log.error("An unknown error occurred: %s" % error)
            success = False
            ret_val = success, e

        return ret_val

    def _check_received(self, msg):
        """If the received message has a field 'answer' that means that the transfer
        successfully completed. Otherwise something went wrong or we received an
        unexpected message."""
        msg_dict = decode_message(msg)
        if "error" in msg_dict:
            error_msg = "A remote error occurred: %s" % msg_dict["error"]
            success = False
            error = TransferError(error_msg)
            log.info(error_msg)
        elif "answer" in msg_dict:
            success = True
            error = None
        else:
            success = False
            error = "Unrecognized message %r" % (msg_dict,)
            log.info(error)
        return success, error

    def stop(self):
        if self.w:
            try:
                self.w.close().addErrback(log.debug)
            except (TransferError, ServerConnectionError, WrongPasswordError) as error:
                # These errors should be already handled previously
                # so here we can safely ignore them
                log.debug("Error: %s", error.type)
            self.w = None


def main(args):
    if not args:
        raise ValueError("You must provide an argument to identify the key")

    def code_generated(result):
        code, _ = result
        print("Discovery info: {}".format(code))
        # Wait for the user without blocking everything
        reactor.callInThread(cancel)

    def cancel():
        input("Press Enter to cancel")
        offer.stop()
        reactor.callFromThread(reactor.stop)

    def received(result):
        success, error_msg = result
        if success:
            print("\nKey successfully sent")
        else:
            print("\nAn error occurred: {}".format(error_msg))
        # We are still waiting for the user to press Enter
        print("Press Enter to exit")

    key = get_usable_keys(pattern=args[0])[0]
    offer = WormholeOffer(key)
    offer.allocate_code().addCallback(code_generated)
    offer.start().addCallback(received)
    print("Offering key: {}".format(key))
    reactor.run()

if __name__ == "__main__":
    import sys
    main(sys.argv[1:])



#!/usr/bin/env python
#    Copyright 2017 Ludovico de Nittis <aasonykk+gnome@gmail.com>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import logging

from twisted.internet.defer import inlineCallbacks
from wormhole.cli.public_relay import RENDEZVOUS_RELAY
from wormhole.errors import WrongPasswordError, LonelyError, TransferError
import wormhole

if __name__ == "__main__":
    from twisted.internet import gireactor
    gireactor.install()
from twisted.internet import reactor

from .gpgmeh import fingerprint_from_keydata
from .i18n import _
from .util import decode_message, encode_message, parse_barcode, mac_verify

log = logging.getLogger(__name__)


class WormholeReceive:
    def __init__(self, code, mac=None, app_id=None):
        self.w = None
        # Check if the given code is a barcode or directly the wormhole code
        parsed = parse_barcode(code).get("WORM", [None])[0]
        if parsed:
            self.code = parsed
        else:
            self.code = code
        if app_id:
            self.app_id = app_id
        else:
            # the following id is needed for interoperability with wormhole cli
            self.app_id = "lothar.com/wormhole/text-or-file-xfer"
        self.mac = mac

    @inlineCallbacks
    def start(self):
        log.info("Wormhole: Trying to receive a message with code: %s", self.code)

        self.stop()
        self.w = wormhole.create(self.app_id, RENDEZVOUS_RELAY, reactor)
        # The following mod is required for Python 2 support
        self.w.set_code("%s" % str(self.code))

        try:
            message = yield self.w.get_message()
            m = decode_message(message)
            key_data = None
            offer = m.get("offer", None)
            if offer:
                key_data = offer.get("message", None)
            if key_data:
                log.info("Message received: %s", key_data)
                if self._is_verified(key_data.encode("utf-8")):
                    log.debug("MAC is valid")
                    success = True
                    message = ""
                    # send a reply with a message ack, this also ensures wormhole cli interoperability
                    reply = {"answer": {"message_ack": "ok"}}
                    reply_encoded = encode_message(reply)
                    self.w.send_message(reply_encoded)
                    ret_val = key_data.encode("utf-8"), success, message
                else:
                    log.warning("The received key has a different MAC")
                    self._reply_error(_("Wrong message authentication code"))
                    ret_val = self._handle_failure(WrongPasswordError())
            else:
                log.info("Unrecognized message: %s", m)
                self._reply_error("Unrecognized message")
                ret_val = self._handle_failure(TransferError())
        except WrongPasswordError as wpe:
            log.info("A wrong password has been used")
            ret_val = self._handle_failure(wpe)
        except LonelyError as le:
            log.info("Closed the connection before we found anyone")
            ret_val = self._handle_failure(le)

        return ret_val

    def _is_verified(self, key_data):
        if self.mac is None:
            # Currently the MAC is not mandatory
            verified = True
        else:
            mac_key = fingerprint_from_keydata(key_data)
            verified = mac_verify(mac_key.encode('ascii'), key_data, self.mac)
        return verified

    def _reply_error(self, error_message):
        reply = {"error": error_message}
        reply_encoded = encode_message(reply)
        self.w.send_message(reply_encoded)

    @inlineCallbacks
    def stop(self):
        if self.w:
            try:
                yield self.w.close()
            except Exception as e:
                print(e)

    @staticmethod
    def _handle_failure(error):
        success = False
        key_data = None
        return key_data, success, type(error)


def main(args):
    log.debug('Running main with args: %s', args)
    if not args:
        raise ValueError("You must provide an argument with the wormhole code")

    @inlineCallbacks
    def receive_key(w_code):
        receive = WormholeReceive(w_code)
        msg_tuple = yield receive.start()
        key_data, success, message = msg_tuple
        if success:
            print("key received:\n")
            print(key_data.decode("utf-8"))
        else:
            print(message)
        # Workaround for the send_message reply (until we find a better solution).
        # If we simply call reactor.stop() the send_message() will never be
        # completed. I think this happens because we are always in the reactor
        # thread and send_message is waiting for a context switch.
        reactor.callLater(1, reactor.stop)

    print("Trying to download the key, please wait")
    code = args[0]
    receive_key(code)
    reactor.run()

if __name__ == "__main__":
    import sys
    main(sys.argv[1:])



#!/usr/bin/env python
#    Copyright 2016 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.
from collections import namedtuple
from datetime import datetime
import logging
import warnings

log = logging.getLogger(__name__)


def to_valid_utf8_string(s, errors='replace', replacement='?'):
    """Takes a string and returns a valid utf8 encodable string

    Not every Python string is utf-8 encodable.
    Take 'fo\\udcf6e\\udce9ba <foo@bma.d>' for example.
    This function replaces undecodable characters with a '?'
    """
    try:
        safe = s.encode('utf-8', errors=errors).decode('utf-8', errors=errors)
    except UnicodeDecodeError:
        # This is the Python 2 way...
        safe = s.decode('utf-8', errors=errors).replace(u"\uFFFD", replacement)
    return safe


def parse_uid(uid, errors='replace'):
    """Parses a GnuPG UID into it's name, comment, and email component
    
    It converts them to strings with the errors paramenter controlling
    how to deal with encoding problems.
    """
    # remove the comment from UID (if it exists)
    com_start = uid.find(b'(')
    if com_start != -1:
        com_end = uid.find(b')')
        uid = uid[:com_start].strip() + uid[com_end+1:].strip()

    # FIXME: Actually parse the comment...
    comment = b""
    comment = comment.decode('utf-8', errors)

    # split into user's name and email
    tokens = uid.split(b'<')
    name = tokens[0].strip()
    name = name.decode('utf-8', errors)
    email = b'unknown'
    if len(tokens) > 1:
        log.debug("Parsing tokens: %r", tokens)
        email = tokens[1].replace(b'>',b'').strip()
    email = email.decode('utf-8', errors)
    
    log.debug("Parsed %r to name (%d): %r", uid, len(name), name)
    return (name, comment, email)


def parse_expiry(value):
    """Takes either a string, an epoch, or a datetime and converts
    it to a datetime.
    If the string is empty (or otherwise evaluates to False)
    then this function returns None, meaning that no expiry has been set.
    An edge case is the epoch value "0".
    """
    if not value:
        expiry = None
    else:
        try:
            expiry = datetime.fromtimestamp(int(value))
        except TypeError:
            expiry = value

    return expiry




class Key(namedtuple("Key", ["expiry", "fingerprint", "uidslist"])):
    "Represents an OpenPGP Key to extent we care about"
    
    log = logging.getLogger(__name__)

    def __new__(cls, expiry, fingerprint, uidslist,
                       *args, **kwargs):
        exp_date = parse_expiry(expiry)
        self = super(Key, cls).__new__(cls, exp_date, fingerprint, uidslist)
        return self

    def __format__(self, arg):
        s  = "{fingerprint}\r\n"
        s += '\r\n'.join(("  {}".format(uid) for uid in self.uidslist))
# This is what original output looks like:
# pub  [unknown] 3072R/1BF98D6D 1336669781 [expiry: 2017-05-09 19:09:41]
#    Fingerprint = FF52 DA33 C025 B1E0 B910  92FC 1C34 19BF 1BF9 8D6D
# uid 1      [unknown] Tobias Mueller <tobias.mueller2@mail.dcu.ie>
# uid 2      [unknown] Tobias Mueller <4tmuelle@informatik.uni-hamburg.de>
# sub   3072R/3B76E8B3 1336669781 [expiry: 2017-05-09 19:09:41]
        return s.format(**self._asdict())

    @property
    def fpr(self):
        """Legacy compatibility, use fingerprint instead.
        However, this is useful for compatibility with gpgme.
        It returns keys with the "fpr" property and we may want
        to be able to run gpgme functions with both their keys and our keys.
        """
        warnings.warn("Legacy fpr, use the fingerprint property",
                      DeprecationWarning)
        return self.fingerprint

    @classmethod
    def from_monkeysign(cls, key):
        "Creates a new Key from an existing monkeysign key"
        log.debug("From mks: %r", key)
        uids = [UID.from_monkeysign(uid) for uid in  key.uidslist]
        expiry = parse_expiry(key.expiry)
        fingerprint = key.fpr
        return cls(expiry, fingerprint, uids)

    @classmethod
    def from_gpgme(cls, key):
        "Creates a new Key from an existing gpgme key"
        uids = [UID.from_gpgme(uid) for uid in  key.uids]
        expiry = parse_expiry(key.subkeys[0].expires)
        fingerprint = key.fpr
        return cls(expiry, fingerprint, uids)



class UID(namedtuple("UID", "expiry uid name comment email")):
    "Represents an OpenPGP UID - at least to the extent we care about it"

    @classmethod
    def from_monkeysign(cls, uid):
        "Creates a new UID from a monkeysign key"
        # We expect to get raw bytes.
        # While RFC4880 demands UTF-8 encoded data,
        # real-life has produced non UTF-8 keys...
        rawuid = to_valid_utf8_string(uid.uid).encode('utf-8')
        log.debug("UidStr (%d): %r", len(rawuid), rawuid)
        name, comment, email = parse_uid(rawuid)
        expiry = parse_expiry(uid.expire)

        return cls(expiry, rawuid.decode('utf-8'),
                   name, comment, email)

    @classmethod
    def from_gpgme(cls, uid):
        "Creates a new UID from a gpgme UID"
        # Weird. I would expect the uid to be raw bytes,
        # because how would gpgme know what encoding to apply?
        # Also, you can have invalid encodings.
        # Turns out, that Python strings can be encoded according to PEP 383
        # which basically encodes invalid bytes as 0xDC80 + byte.
        # That's the "surrogateescape" error handler available in Python 3.
        # Here, we don't care about that, though. We are in the user facing
        # abstraction for a UID. As such, we ensure that it can be rendered.
        # So we take the string we get from gpgme and try to convert it to
        # to utf-8 bytes.
        log.debug("UID from gpgme: %r", uid.uid)
        rawuid = to_valid_utf8_string(uid.uid)
        name = to_valid_utf8_string(uid.name)
        comment = '' # FIXME: uid.comment
        email = to_valid_utf8_string(uid.email)
        expiry = None  #  FIXME: Maybe UIDs don't expire themselves but via the binding signature

        return cls(expiry, rawuid, name, comment, email)



#!/usr/bin/env python
#    Copyright 2016 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import signal
import sys
import argparse
import logging
import os


if __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')

from .__init__ import __version__
from .gpgmeh import get_usable_keys, get_public_key_data
from .i18n import _
from .util import mac_generate, format_fingerprint
from . import Keyserver

log = logging.getLogger(__name__)


class AvahiHTTPOffer:
    "Spawns a local HTTP daemon and announces it via Avahi"
    def __init__(self, key):
        self.key = key
        self.fingerprint = fingerprint = key.fingerprint
        self.keydata = keydata = get_public_key_data(fingerprint)
        self.keyserver = Keyserver.ServeKeyThread(keydata, fingerprint)
        self.mac = mac_generate(fingerprint.encode('ascii'), keydata)

    def allocate_code(self):
        """Returns the information necessary to discover the key through Avahi"""
        fingerprint = self.fingerprint.upper()
        mac = self.mac.upper()
        discovery_info = 'OPENPGP4FPR:{0}#MAC={1}'.format(
                                fingerprint, mac)
        return format_fingerprint(self.key.fingerprint), discovery_info

    def start(self):
        """Starts offering the key"""
        log.info("Requesting to start")
        self.keyserver.start()


    def stop(self):
        "Stops offering the key"
        log.info("Requesting to shutdown")
        self.keyserver.shutdown()


def main(args):
    if not args:
        raise ValueError("You must provide an argument to identify the key")

    key = get_usable_keys(pattern=args[0])[0]
    offer = AvahiHTTPOffer(key)
    discovery_info = offer.allocate_code()
    print (_("Offering key: {}").format(key))
    print (_("Discovery info: {}").format(discovery_info))
    offer.start()
    print (_("Press Enter to stop"))
    try: input_ = raw_input
    except NameError: input_ = input
    input_("Press Enter to stop")
    offer.stop()


if __name__ == "__main__":
    import sys
    main(sys.argv[1:])



import logging

from twisted.internet import threads
from twisted.internet.defer import inlineCallbacks
from wormhole.errors import LonelyError

from .wormholereceive import WormholeReceive
from .avahidiscovery import AvahiKeysignDiscoveryWithMac
from .util import is_code_complete, parse_barcode
try:
    from .bluetoothreceive import BluetoothReceive
except ImportError:
    BluetoothReceive = None

log = logging.getLogger(__name__)


class Discover:
    def __init__(self, userdata, discovery, app_id=None):
        # if the userdata is a qr code we extract the wormhole and bluetooth codes
        self.worm_code = parse_barcode(userdata).get("WORM", [None])[0]
        self.bt_code = parse_barcode(userdata).get("BT", [None])[0]
        self.bt_port = parse_barcode(userdata).get("PT", [None])[0]
        if self.bt_port:
            self.bt_port = int(self.bt_port)
        self.mac = parse_barcode(userdata).get("MAC", [None])[0]
        # check if userdata is a valid wormhole code
        if is_code_complete(userdata):
            self.worm_code = userdata
        self.userdata = userdata
        self.app_id = app_id
        if discovery:
            self.discovery = discovery
        else:
            self.discovery = AvahiKeysignDiscoveryWithMac()
        self.worm = None
        self.bt = None
        self.stopped = False

    @inlineCallbacks
    def start(self):
        # First we try Avahi, if it fails we fallback to Bluetooth and lastly
        # Wormhole, because the receiver may be able to use only one of them
        success = False
        message = LonelyError
        log.info("Trying to use this code with Avahi: %s", self.userdata)
        
        try:
            key_data = yield threads.deferToThread(self.discovery.find_key, self.userdata)
        except ValueError as e:
            key_data = None
            success = False
            message = "Error downloading key, maybe it has been altered in transit"
            log.warning(message, exc_info=e)
        else:
            # Actually.. key_data can very well be None as an indication of failure. We might change that API to throw.
            log.debug("We may have found a key: %r", key_data)

        if key_data:
            success = True
            message = ""
        else:
            if self.bt_code and BluetoothReceive and not self.stopped:
                # We try Bluetooth, if we have it
                log.info("Trying to connect to %s with Bluetooth", self.bt_code)
                # We try to see if Bluetooth was imported,
                # else we log an event of missing Pybluez.
                try:
                    self.bt = BluetoothReceive(self.bt_port)
                    msg_tuple = yield self.bt.find_key(self.bt_code, self.mac)
                except TypeError as e:
                    key_data = None
                    success = False
                    log.exception("Pybluez may be missing.")
                else:
                    key_data, success, message = msg_tuple
                    if key_data:
                        # If we found the key
                        log.debug("Found the key via bluetooth: %r", key_data[:32])

            if not key_data and self.worm_code and not self.stopped:
                # We try the wormhole code, if we have it
                log.info("Trying to use this code with Wormhole: %s", self.worm_code)
                self.worm = WormholeReceive(self.worm_code, self.mac)
                msg_tuple = yield self.worm.start()
                key_data, success, message = msg_tuple

        if self.stopped:
            key_data = None
            success = False
            # We use the LonelyError in a similar way as Wormhole does.
            # That is to indicate that the connection as been stopped before a transfer.
            message = LonelyError

        log.debug("Returning key: %r, success: %r, message: %r",
                  key_data, success, message)
        return key_data, success, message

    def stop(self):
        self.stopped = True
        # WormholeReceive needs to be stopped because right now after the 'start()'
        # it continues trying to connect until it does or we stop it.
        if self.worm:
            self.worm.stop()
        if self.bt:
            self.bt.stop()



import logging
import select
import socket

if __name__ == "__main__":
    import gi
    gi.require_version('Gtk', '4.0')
    from twisted.internet import gireactor
    gireactor.install()
    from twisted.internet import reactor
from twisted.internet import threads
from twisted.internet.defer import inlineCallbacks

if __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    import os
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')

from .gpgmeh import fingerprint_from_keydata
from .i18n import _
from .util import mac_verify

log = logging.getLogger(__name__)


class BluetoothError(OSError):
    pass


class BluetoothReceive:
    def __init__(self, port=3, size=1024):
        self.port = port
        self.size = size
        self.client_socket = None
        self.stopped = False

    @inlineCallbacks
    def find_key(self, bt_mac, mac):
        self.client_socket = socket.socket(
                socket.AF_BLUETOOTH,
                socket.SOCK_STREAM,
                socket.BTPROTO_RFCOMM)
        message = b""
        success = False
        ret_val = None
        try:
            self.client_socket.setblocking(False)
            try:
                log.info("Trying to connect to %s port %s", bt_mac, self.port)
                self.client_socket.connect((bt_mac, self.port))
            except OSError as be:
                if be.errno == 115: #"(115, 'Operation now in progress')":
                    pass
                else:
                    log.exception("BT connection (%d)", be.errno)
                    raise be
            while not self.stopped and not success:
                r, w, e = yield threads.deferToThread(select.select, [self.client_socket], [], [], 0.5)
                if r:
                    log.info("Connection established")
                    self.client_socket.setblocking(True)
                    success = True
                    # try to receive until the sender closes the connection
                    try:
                        while True:
                            part_message = self.client_socket.recv(self.size)
                            log.debug("Read %d bytes: %r", len(part_message), part_message)
                            message += part_message
                    except ConnectionRefusedError as be:
                        log.exception("Connection Refused!")
                    except OSError as be:
                        log.exception("BT recv (%d)", be.errno)
                        if be.errno == 104: # "(104, 'Connection reset by peer')":
                            log.info("Bluetooth connection closed, let's check if we downloaded the key")
                        else:
                            raise be
            mac_key = fingerprint_from_keydata(message)
            verified = None
            if mac:
                verified = mac_verify(mac_key.encode('ascii'), message, mac)
            if verified:
                success = True
            else:
                log.info("MAC validation failed: %r", verified)
                success = False
                message = b""
            ret_val = message.decode("utf-8"), success, None
        except ConnectionRefusedError as be:
            log.info("CR")
            ret_val = message.decode("utf-8"), success, None
        except OSError as be:
            log.exception("BT conn (%d)", be.errno)
            if be.errno == 16: # "(16, 'Device or resource busy')":
                log.info("Probably has been provided a partial bt mac")
            elif be.errno == 111: # "(111, 'Connection refused')":
                log.info("The sender refused our connection attempt")
            elif be.errno == 112: #  "(112, 'Host is down')":
                log.info("The sender's Bluetooth is not available")
            elif be.errno == 113: # "(113, 'No route to host')":
                log.info("An error occurred with Bluetooth, if present probably the device is not powered")
            else:
                log.info("An unknown bt error occurred: %s" % be.errno)

            key_data = None
            success = False
            ret_val = key_data, success, be
        except Exception as e:
            log.error("An error occurred connecting or receiving: %s" % e)
            key_data = None
            success = False
            ret_val = key_data, success, e

        if self.client_socket:
            self.client_socket.close()
        return ret_val

    def stop(self):
        self.stopped = True
        if self.client_socket:
            try:
                self.client_socket.shutdown(socket.SHUT_RDWR)
                self.client_socket.close()
            except OSError as be:
                if be.errno == 9: #"(9, 'Bad file descriptor')":
                    log.info("The old Bluetooth connection was already closed")
                else:
                    log.exception("An unknown bt error occurred")


def main(args):
    log.debug('Running main with args: %s', args)
    if not len(args) == 3:
        raise ValueError("You must provide three arguments: bluetooth code, hmac and port")

    def _received(result):
        key_data, success, error_message = result
        if success:
            print(key_data)
        else:
            print(error_message)

        reactor.callFromThread(reactor.stop)

    print(_("Trying to download the key, please wait"))
    bt_mac = args[0]
    hmac = args[1]
    port = int(args[2])
    receive = BluetoothReceive(port)
    d = receive.find_key(bt_mac, hmac)
    d.addCallback(_received)
    reactor.run()

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    import sys
    main(sys.argv[1:])



#!/usr/bin/env python
#    Copyright 2014 Tobias Mueller <muelli@cryptobitch.de>
#    Copyright 2014 Andrei Macavei <andrei.macavei89@gmail.com>
#    Copyright 2015 Jody Hansen <jobediah.hansen@gmail.com>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

try:
    from http.server import BaseHTTPRequestHandler, HTTPServer
    from socketserver import ThreadingMixIn
except ImportError:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
    from SocketServer import ThreadingMixIn
import logging
import os
import socket
from threading import Thread

# This is probably really bad...  But doing relative imports only
# works for modules.  However, I want to be able to call this Keyserver.py
# for testing purposes.
if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')

from .__init__ import __version__
from .network.AvahiPublisher import AvahiPublisher

from .gpgmeh import fingerprint_from_keydata

log = logging.getLogger(__name__)

class KeyRequestHandlerBase(BaseHTTPRequestHandler):
    '''This is the "base class" which needs to be given access
    to the key to be served. So you will not use this class,
    but create a use one inheriting from this class. The subclass
    must also define a keydata field.
    '''
    server_version = 'GNOME-Keysign/' + '%s' % __version__

    # As per RFC 2015 Section 7
    # https://tools.ietf.org/html/rfc2015#section-7
    ctype = 'application/pgp-keys'

    def do_GET(self):
        f = self.send_head(self.keydata)
        self.wfile.write(self.keydata)

    def send_head(self, keydata=None):
        kd = keydata if keydata else self.keydata
        self.send_response(200)
        self.send_header('Content-Type', self.ctype)
        self.send_header('Content-Length', len(kd))
        self.end_headers()
        return kd

class ThreadedKeyserver(ThreadingMixIn, HTTPServer):
    '''The keyserver in a threaded fashion'''
    address_family = socket.AF_INET6

    def __init__(self, server_address, *args, **kwargs):
        if issubclass(self.__class__, object):
            super(ThreadedKeyserver, self).__init__(server_address, *args, **kwargs)
        else:
            HTTPServer.__init__(self, server_address, *args, **kwargs)
            # WTF? There is no __init__..?
            # ThreadingMixIn.__init__(self, server_address, *args, **kwargs)

        def server_bind(self):
            # Override this method to be sure v6only is false: we want to
            # listen to both IPv4 and IPv6!
            self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, False)
            HTTPServer.server_bind(self)



class ServeKeyThread(Thread):
    '''Serves requests and manages the server in separates threads.
    You can create an object and call start() to let it run.
    If you want to stop serving, call shutdown().
    '''

    def __init__(self, data, fpr, port=9001, *args, **kwargs):
        '''Initializes the server to serve the data'''
        self.keydata = data
        self.fpr = fpr
        self.port = port
        super(ServeKeyThread, self).__init__(*args, **kwargs)
        self.daemon = True
        self.httpd = None


    def start(self, data=None, fpr=None, port=None, *args, **kwargs):
        '''This is run in the same thread as the caller.
        This calls run() in a separate thread.
        In order to resolve DBus issues, most things
        are done here.
        However, you probably need to start
        dbus.mainloop.glib.DBusGMainLoop (set_as_default=True)
        in order for this work.
        '''

        port = port or self.port or 9001
        fpr = fpr or self.fpr

        tries = 10

        kd = data if data else self.keydata

        class KeyRequestHandler(KeyRequestHandlerBase):
            '''You will need to create this during runtime'''
            keydata = kd
        HandlerClass = KeyRequestHandler

        for port_i in (port + p for p in range(tries)):
            try:
                log.info('Trying port %d', port_i)
                server_address = ('', port_i)
                self.httpd = ThreadedKeyserver(server_address, HandlerClass, **kwargs)

                ###
                # This is a bit of a hack, it really should be
                # in some lower layer, such as the place were
                # the socket is created and listen()ed on.
                service_txt = {
                    'fingerprint': fpr,
                    'version': __version__,
                }


                log.info('Requesting Avahi with txt: %s', service_txt)

                self.avahi_publisher = ap = AvahiPublisher(
                    service_port = port_i,
                    service_name = 'HTTP Keyserver %s' % fpr,
                    service_txt = service_txt,
                    # self.keydata is too big for Avahi; it crashes
                    service_type = '_gnome-keysign._tcp',
                )
                log.info('Trying to add Avahi Service')
                ap.add_service()

            except socket.error as value:
                errno = value.errno
                if errno == 10054 or errno == 32:
                    # This seems to be harmless
                    break
            else:
                break

            finally:
                pass


        super(ServeKeyThread, self).start(*args, **kwargs)


    def serve_key(self, poll_interval=0.15):
        '''An HTTPd is started and being put to serve_forever.
        You need to call shutdown() in order to stop
        serving.
        '''

        #sa = self.httpd.socket.getsockname()
        try:
            log.info('Serving now on %s, this is probably blocking...',
                     self.httpd.socket.getsockname())
            self.httpd.serve_forever(poll_interval=poll_interval)
        finally:
            log.info('finished serving')
            #httpd.dispose()



    def run(self):
        '''This is being run by Thread in a separate thread
        after you call start()'''
        self.serve_key()


    def shutdown(self):
        '''Sends shutdown to the underlying httpd'''
        log.info("Removing Avahi Service")
        self.avahi_publisher.remove_service()
        log.info("Shutting down httpd %r", self.httpd)
        self.httpd.shutdown()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    import dbus, time
    dbus.mainloop.glib.DBusGMainLoop (set_as_default=True)
    def stop_thread(t, seconds=5):
        log.info('Sleeping %d seconds, then stopping', seconds)
        time.sleep(seconds)
        t.shutdown()

    import sys
    if len(sys.argv) >= 2:
        fname = sys.argv[1]
        KEYDATA = open(fname, 'rb').read()
        # FIXME: Someone needs to determine the fingerprint
        #        of the data just read
        fpr = fingerprint_from_keydata(KEYDATA)
    else:
        KEYDATA = b'Example data'
        fpr = ''.join('F289 F7BA 977D F414 3AE9  FDFB F70A 0290 6C30 1813'.split())

    if len(sys.argv) >= 3:
        timeout = int(sys.argv[2])
    else:
        timeout = 5

    t = ServeKeyThread(KEYDATA, fpr)
    stop_t = Thread(target=stop_thread, args=(t,timeout))
    stop_t.daemon = True
    t.start()
    stop_t.start()
    while True:
        log.info('joining stop %s', stop_t.isAlive())
        stop_t.join(1)
        log.info('joining t %s', t.isAlive())
        t.join(1)
        if not t.isAlive() or not stop_t.isAlive():
            break
    log.warn('Last line')



#!/usr/bin/env python
# encoding: utf-8
#    Copyright 2014 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
import signal
import sys
import argparse
import logging

import gi
gi.require_version('Gtk', '4.0')
from gi.repository import Gtk, GLib
from gi.repository import GObject

from .gpgmeh import get_usable_secret_keys, get_usable_keys

# These are relative imports
from .__init__ import __version__

log = logging.getLogger(__name__)


class KeysPage(Gtk.Box):
    '''This represents a list of keys with the option for the user
    to select one key to proceed.
    
    This class emits a `key-selection-changed' signal when the user
    initially selects a key such that it is highlighted.
    
    The `key-selected' signal is emitted when the user commits
    to a key, i.e. by pressing a designated button to make his
    selection public.
    '''
    __gsignals__ = {
        str('key-selected'): (GObject.SIGNAL_RUN_LAST, None,
                         # Hm, this is a str for now, but ideally
                         # it'd be the full key object
                         (str,)),
        str('key-selection-changed'): (GObject.SIGNAL_RUN_LAST, None,
                         # Hm, this is a str for now, but ideally
                         # it'd be the full key object
                         (str,)),
    }

    def __init__(self, show_public_keys=False):
        '''Sets the widget up.
        
        The show_public_keys parameter is meant for development
        purposes only.  If set to True, the widget will show
        the public keys, too.  Otherwise, secret keys are shown.
        '''
        super(KeysPage, self).__init__(orientation=Gtk.Orientation.VERTICAL)

        # set up the list store to be filled up with user's gpg keys
        # Note that other functions expect a certain structure to
        # this ListStore, e.g. when parsing the selection of the
        # TreeView, i.e. in get_items_from_selection.
        self.store = Gtk.ListStore(str, str, str)
        #                       name, email, fingerprint

        keys = get_usable_secret_keys()
        keys += get_usable_keys() if show_public_keys else []
        for key in keys:
            uidslist = key.uidslist #UIDs: Real Name (Comment) <email@address>
            fingerprint = key.fingerprint

            for uid in uidslist:
                self.store.append((uid.name, uid.email, fingerprint))

        if len(self.store) == 0:
            lbl = Gtk.Label(label="You don't have a private key")
            lbl.set_hexpand(True)
            lbl.set_vexpand(True)
            self.append(lbl)
        else:
            # create the tree view
            self.treeView = Gtk.TreeView(model=self.store)
            # setup 'Name' column
            nameRenderer = Gtk.CellRendererText()
            nameColumn = Gtk.TreeViewColumn("Name", nameRenderer, text=0)

            # setup 'Email' column
            emailRenderer = Gtk.CellRendererText()
            emailColumn = Gtk.TreeViewColumn("Email", emailRenderer, text=1)

            ## setup 'Fingerprint' column
            # keyRenderer = Gtk.CellRendererText()
            # keyColumn = Gtk.TreeViewColumn("Fingerprint", keyRenderer, text=2)

            self.treeView.append_column(nameColumn)
            self.treeView.append_column(emailColumn)
            # self.treeView.append_column(keyColumn)

            self.treeView.connect('row-activated', self.on_row_activated)

            # make the tree view resposive to single click selection
            self.treeView.get_selection().connect('changed', self.on_selection_changed)

            # make the tree view scrollable
            self.scrolled_window = Gtk.ScrolledWindow()
            self.scrolled_window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
            self.scrolled_window.set_child(self.treeView)
            self.scrolled_window.set_min_content_height(200)

            #self.pack_start(self.scrolled_window, True, True, 0)

            self.hpane = Gtk.Paned(orientation=Gtk.Orientation.HORIZONTAL)
            self.hpane.set_start_child(self.scrolled_window)
            self.right_pane = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            right_label = Gtk.Label(label='Select key on the left')
            self.right_pane.append(right_label)
            self.hpane.set_end_child(self.right_pane)

            self.hpane.set_hexpand(True)
            self.hpane.set_vexpand(True)
            self.append(self.hpane)


    # We could make it a @staticmethod, but the returned items
    # are bound to the model, anyway.  So it probably doesn't
    # make much sense to have a static function, anyway.
    def get_items_from_selection(self, selection=None):
        '''Returns the elements in the ListStore for the given selection'''
        s = selection or self.treeView.get_selection()
        model, paths = s.get_selected_rows()
        name = email = fingerprint = None
        for path in paths:
            iterator = model.get_iter(path)
            (name, email, fingerprint) = model.get(iterator, 0, 1, 2)
            break

        return (name, email, fingerprint)


    def on_selection_changed(self, selection, *args):
        log.debug('Selected new TreeView item %s = %s', selection, args)
        
        name, email, fingerprint = \
            self.get_items_from_selection(selection)[:3]
        
        self.emit('key-selection-changed', fingerprint)
        
        # FIXME: We'd rather want to get the key object
        # (or its representation) from the model, not by querying again
        key = next(iter(get_usable_keys(pattern=fingerprint)))
        exp_date = key.expiry

        if exp_date is None:
            expiry = "No expiration date"
        else:
            expiry = "{:%Y-%m-%d %H:%M:%S}".format(exp_date)

        pane = self.right_pane
        child = pane.get_first_child()
        while child:
            next_child = child.get_next_sibling()
            pane.remove(child)
            child = next_child

        ctx = {'keyid':fingerprint[-8:], 'expiry':expiry,
               'sigs':'', 'fingerprint':fingerprint}
        keyid_label = Gtk.Label(label='Key {keyid}'.format(**ctx))
        expiration_label = Gtk.Label(label='Expires: {expiry}'.format(**ctx))
        #signatures_label = Gtk.Label(label='{sigs} signatures'.format(**ctx))
        publish_button = Gtk.Button(label='Go ahead!'.format(**ctx))
        publish_button.connect('clicked', self.on_publish_button_clicked, key)
        for w in (keyid_label
                  , expiration_label
                  #, signatures_label
                  , publish_button
                  ):
            pane.append(w)


    def on_row_activated(self, treeview, tree_path, column):
        '''A callback for when the user "activated" a row,
        e.g. by double-clicking an entry.
        
        It emits the key-selected signal.
        '''
        # We just hijack the existing function.
        # I'm sure we could get the required information out of
        # the tree_path and column, but I don't know how.
        name, email, fingerprint = \
            self.get_items_from_selection()[:3]
        self.emit('key-selected', fingerprint)


    def on_publish_button_clicked(self, button, key, *args):
        '''Callback for when the user has expressed their wish
        to publish a key on the network.  It will emit a "key-selected"
        signal with the ID of the selected key.'''
        log.debug('Clicked publish for key (%s) %s (%s)', type(key), key, args)
        fingerprint = key.fingerprint
        self.emit('key-selected', fingerprint)




class Keys(Gtk.Application):
    """A widget which displays keys in a user's Keyring.
    
    Once the user has selected a key, the key-selected
    signal will be thrown.
    """
    def __init__(self, *args, **kwargs):
        #super(Keys, self).__init__(*args, **kwargs)
        Gtk.Application.__init__(
            self, application_id="org.gnome.keysign.keys")
        self.connect("activate", self.on_activate)
        self.connect("startup", self.on_startup)

        self.log = logging.getLogger(__name__)

        self.keys_page = KeysPage()
        self.keys_page.connect('key-selection-changed',
            self.on_key_selection_changed)
        self.keys_page.connect('key-selected', self.on_key_selected)


    def on_quit(self, app, param=None):
        self.quit()


    def on_startup(self, app):
        self.log.info("Startup")
        self.window = Gtk.ApplicationWindow(application=app)
        self.window.set_title ("Keysign - Keys")
        self.window.set_child(self.keys_page)


    def on_activate(self, app):
        self.log.info("Activate!")
        #self.window = Gtk.ApplicationWindow(application=app)

        # In case the user runs the application a second time,
        # we raise the existing window.
        self.window.present()


    def on_key_selection_changed(self, button, key):
        """This is the connected to the KeysPage's key-selection-changed
        signal
        
        As a user of that widget, you would show more details
        in the GUI or prepare for a final commitment by the user.
        """
        self.log.info('Selection changed to: %s', key)


    def on_key_selected(self, button, fpr):
        """This is the connected to the KeysPage's key-selected signal
        
        As a user of that widget, you would enable buttons or proceed
        with the GUI.
        """
        self.log.info('User committed to a key! %s', fpr)

                                                
def parse_command_line(argv):
    """Parse command line argument. See -h option

    :param argv: arguments on the command line must include caller file name.
    """
    formatter_class = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(description='Auxiliary helper program '+
                                                 'display keys',
                                     formatter_class=formatter_class)
    parser.add_argument("--version", action="version",
                        version="%(prog)s {}".format(__version__))
    parser.add_argument("-v", "--verbose", dest="verbose_count",
                        action="count", default=0,
                        help="increases log verbosity for each occurence.")
    #parser.add_argument('-o', metavar="output",
    #                    type=argparse.FileType('w'), default=sys.stdout,
    #                    help="redirect output to a file")
    #parser.add_argument('input', metavar="input",
    ## nargs='+', # argparse.REMAINDER,
    #help="input if any...")
    arguments = parser.parse_args(argv[1:])
    # Sets log level to WARN going more verbose for each new -v.
    log.setLevel(max(3 - arguments.verbose_count, 0) * 10)
    return arguments


def main(args=sys.argv):
    """This is an example program of how to use the Keys widget"""
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
                        format='%(name)s (%(levelname)s): %(message)s')
    try:
        arguments = parse_command_line(args)
        
        app = Keys()
        try:
            GLib.unix_signal_add_full(GLib.PRIORITY_HIGH, signal.SIGINT, lambda *args : app.quit(), None)
        except AttributeError:
            pass
    
        exit_status = app.run(None)
    
        return exit_status

        
    finally:
        logging.shutdown()

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

import logging, sys, os
import gi
gi.require_version('Gtk', '4.0')
from gi.repository import Gtk

thisdir = os.path.dirname(os.path.realpath(__file__))
# Add parent directory to path so we can import keysign module
sys.path.insert(0, os.path.dirname(thisdir))
from keysign.keyconfirm import PreSignApp

class TestImportApp(PreSignApp):
    def on_activate(self, app):
        super().on_activate(app)
        # Force the "certifications produced" infobar to be visible
        # This contains our new Import button!
        self.psw.infobar_success.set_visible(True)

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    app = TestImportApp()
    app.run(sys.argv[1:])



#!/usr/bin/env python
#    Copyright 2017 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re
import os
import signal
import sys

import gi
gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')
from gi.repository import Gtk, GLib, Adw, Gio
gi.require_version('Gst', '1.0')
from gi.repository import Gst
from gi.repository import Gdk
try:
    from twisted.internet import gireactor
    gireactor.install()
except Exception:
    pass

from twisted.internet import reactor


if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')


from .avahioffer import AvahiHTTPOffer
from .avahidiscovery import AvahiKeysignDiscoveryWithMac
from .errors import NoAvahiDbus
from .keyconfirm import PreSignWidget
from .keyfprscan import KeyFprScanWidget
from .keylistwidget import KeyListWidget
from .KeyPresent import KeyPresentWidget
from .gpgmeh import openpgpkey_from_data
from . import gpgmeh
from .i18n import _
from .receive import ReceiveApp
from .send import SendApp
from . import gtkexcepthook


log = logging.getLogger(__name__)

def remove_whitespace(s):
    cleaned = re.sub(r'[\s+]', '', s)
    return cleaned



class PswMappingReceiveApp(ReceiveApp):
    """A simple extension to the existing Receive class
    to connect to the PreSignWidget's mapped signal (or
    an emulation thereof.  This is a bit of a hack, but by
    having pushed common receive functionality in the ReceiveApp
    class, we do not necessarily control anymore when the 
    PreSignWidget is created let alone connect to the map signal
    in time.
    """
    def __init__(self, mapped_func, builder=None):
        # ReceiveApp, in Python 2, is an old style object
        ReceiveApp.__init__(self, builder=builder)
        self.func = mapped_func
        
    def on_keydata_downloaded(self, *args, **kwargs):
        ReceiveApp.on_keydata_downloaded(self, *args, **kwargs)
        psw = self.psw
        psw.connect('map', self.func)
        if psw.get_mapped():
            self.func(psw)

    def get_toplevel(self):
        """We're continuing to hack around
        If we ever want to run a dialog, say a FileSave dialog, then we need to provide
        the top level window of that widget.
        """
        return self.psw.get_toplevel()


class KeysignApp(Adw.Application):
    def __init__(self, *args, **kwargs):
        super(KeysignApp, self).__init__(*args, **kwargs)
        self.connect('activate', self.on_activate)

        self.send_stack = None
        self.receive_stack = None
        self.send_receive_stack = None
        self.header_button_handler_id = None
        self.pre_sign_widget = None

    def on_activate(self, app):
        ui_file_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "app4.ui")
        appwindow = 'applicationwindow1'
        builder = Gtk.Builder()
        builder.add_objects_from_file(ui_file_path, [appwindow])
        window = builder.get_object(appwindow)
        window.set_title("GNOME Keysign")
        window.connect("close-request", self.on_delete_window)

        def on_realize(win):
            surface = win.get_surface()
            try:
                from gi.repository import GdkWayland
                if isinstance(surface, GdkWayland.WaylandToplevel):
                    def on_handle_exported(toplevel, handle, *args):
                        win.portal_handle = f"wayland:{handle}"
                    surface.export_handle(on_handle_exported)
            except Exception:
                pass
        window.connect("realize", on_realize)
        self.headerbar = window.get_titlebar()
        self.header_button = builder.get_object("back_refresh_button")
        self.header_button.connect('clicked', self.on_header_button_clicked)
        self.internet_toggle = builder.get_object("internet_toggle")
        self.internet_toggle.connect("toggled", self.on_toggle_clicked)

        sw = builder.get_object('stackswitcher1')
        # FIXME: I want to be able to press Alt+S and Alt+R respectively
        # to switch the stack pages to Send and Receive.
        # It's possible when using the Gtk Inspector and modify the
        # Switcher's children (ToggleButton and Label) to "use-underscore".
        # but it must be possible to do programmatically.
        # sw.get_children()
        self.stack_switcher = sw

        self.send_receive_stack = builder.get_object("send_receive_stack")
        self.send_receive_stack.connect('notify::visible-child',
            self.on_sr_stack_switch)


        ## Load Send part
        self.send = SendApp()
        ss = self.send.stack
        p = ss.get_parent()
        if p:
            p.remove(ss)
        ss.connect('notify::visible-child', self.on_send_stack_switch)
        ss.connect('map', self.on_send_stack_mapped)
        self.send_stack = ss
        ## End of loading send part


        # Load Receive part
        try:
            self.receive = PswMappingReceiveApp(self.on_presign_mapped)
        except NoAvahiDbus as de:
            log.info("Probably Avahi needs to be manually started: %s", de)
            ui_file_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "dialog_avahi4.ui")
            appwindow = 'dialog_avahi'
            builder = Gtk.Builder()
            builder.add_objects_from_file(ui_file_path, [appwindow])
            window.set_title("GNOME Keysign")
            ok_button = builder.get_object("avahi_ok")
            ok_button.connect('clicked', self.on_delete_window)
            window = builder.get_object(appwindow)
        else:
            rs = self.receive.stack

            rs.connect('notify::visible-child',
                self.on_receive_stack_switch)


            scanner = self.receive.scanner
            scanner.connect("map", self.on_scanner_mapped)
            self.receive_stack = rs


            # Hm. Leaving comments for translators does not seem to work
            self.send_receive_stack.add_titled(self.send_stack,
                "send_stack", _("Send"))
            self.send_receive_stack.add_titled(rs,
                "receive_stack", _("Receive"))

            # Actions and accelerators to switch between Send and Receive tabs using Alt+S and Alt+R
            send_action = Gio.SimpleAction.new("switch-to-send", None)
            send_action.connect("activate", lambda action, parameter: self.send_receive_stack.set_visible_child_name("send_stack"))
            self.add_action(send_action)

            receive_action = Gio.SimpleAction.new("switch-to-receive", None)
            receive_action.connect("activate", lambda action, parameter: self.send_receive_stack.set_visible_child_name("receive_stack"))
            self.add_action(receive_action)

            self.set_accels_for_action("app.switch-to-send", ["<Alt>s"])
            self.set_accels_for_action("app.switch-to-receive", ["<Alt>r"])

        window.present()
        self.add_window(window)

    def run(self, args=[]):
        super(KeysignApp, self).run()

    @staticmethod
    def on_delete_window(*args):
        reactor.callFromThread(reactor.stop)

    def on_sr_stack_switch(self, stack, *args):
        log.debug("Switched Stack! %r", args)
        #self.update_header_button()

    def on_send_stack_switch(self, stack, *args):
        log.debug("Switched Send Stack! %r", args)
        current = self.send.stack.get_visible_child()
        if current == self.send.klw:
            log.debug("Key list page now visible")
            self.on_keylist_mapped(self.send.klw)
        elif current == self.send.kpw:
            log.debug("Key present page now visible")
            self.on_keypresent_mapped(self.send.kpw)
        elif current == self.send.rb:
            log.debug("Result page now visible")
            self.on_resultbox_mapped(self.send.rb)
        else:
            log.error("An unexpected page is now visible: %r", current)

    def on_receive_stack_switch(self, stack, *args):
        log.debug("Switched Receive Stack! %r", args)
        #self.update_header_button()

    def on_send_header_button_clicked(self, button, *args):
        # Here we assume that there are only two places where
        # we could have possibly pressed this button, i.e.
        # from the keypresentwidget or the result page
        log.debug("Send Headerbutton %r clicked! %r", button, args)
        current = self.send.stack.get_visible_child()
        klw = self.send.klw
        kpw = self.send.kpw
        rb = self.send.rb
        # If we are in the keypresentwidget
        if current == kpw:
            self.send_stack.set_visible_child(klw)
            self.send.deactivate()
        # If we are in the result page
        elif current == rb:
            self.send_stack.remove(current)
            self.send.set_saved_child_visible()
            self.send.on_key_activated(None, self.send.key)
        else:
            log.error("Header button pressed in an unexpected page: %s", current)

    def on_receive_header_button_clicked(self, button, *args):
        # Here we assume that there is only one place where
        # we could have possibly pressed this button, i.e.
        # from the presignwidget.
        log.debug("Receive Headerbutton %r clicked! %r", button, args)
        self.receive_stack.set_visible_child_name("scanner")

    def on_header_button_clicked(self, button, *args):
        log.debug("Headerbutton %r clicked! %r", button, args)
        # We have 2 children in the top level stack: send and receive.
        # In the send stack, we currently have two children.
        # In the receive stack, we have at least three.
        visible_child = self.send_receive_stack.get_visible_child()
        if not visible_child:
            return
        if visible_child == self.send_stack:
            return self.on_send_header_button_clicked(button, *args)
        elif visible_child == self.receive_stack:
            return self.on_receive_header_button_clicked(button, *args)
        else:
            raise RuntimeError("We expected either send or receive stack "
                "but got %r" % visible_child)

    def on_toggle_clicked(self, toggle):
        log.info("Internet toggled to: %s", toggle.get_active())
        self.send.set_internet_option(toggle.get_active())

    def on_resultbox_mapped(self, rb):
        log.debug("Resultbox becomes visible!")
        self.header_button.set_sensitive(True)
        self.header_button.set_icon_name("go-previous")
        self.internet_toggle.hide()

    def on_keylist_mapped(self, keylistwidget):
        log.debug("Keylist becomes visible!")
        self.header_button.set_icon_name("view-refresh")
        # We don't support refreshing for now.
        self.header_button.set_sensitive(False)
        self.internet_toggle.show()

    def on_send_stack_mapped(self, stack):
        log.debug("send stack becomes visible!")
        # Adjust the top bar buttons
        self.on_send_stack_switch(stack)

    def on_keypresent_mapped(self, kpw):
        log.debug("keypresent becomes visible!")
        self.header_button.set_sensitive(True)
        self.header_button.set_icon_name("go-previous")
        self.internet_toggle.hide()

    def on_scanner_mapped(self, scanner):
        log.debug("scanner becomes visible!")
        self.header_button.set_sensitive(False)
        self.header_button.set_icon_name("go-previous")
        self.internet_toggle.hide()

    def on_presign_mapped(self, psw):
        log.debug("presign becomes visible!")
        self.header_button.set_sensitive(True)
        self.header_button.set_icon_name("go-previous")
        self.internet_toggle.hide()


def main(args=[]):
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(name)s (%(levelname)s): %(message)s')
    log = logging.getLogger(__name__)
    log.debug('Running main with args: %s', args)
    if not args:
        args = []
    Gst.init(None)

    Adw.init()
    app = KeysignApp(application_id="org.gnome.Keysign")
    try:
        GLib.unix_signal_add_full(GLib.PRIORITY_HIGH, signal.SIGINT,
                                  lambda *args: reactor.callFromThread(reactor.stop), None)
    except AttributeError:
        pass
    reactor.registerGApplication(app)
    reactor.run()

if __name__ == '__main__':
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
            format='%(name)s (%(levelname)s): %(message)s')
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python
#    Copyright 2016 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import hmac
import json
import logging
import mailbox
import os
import shutil
from subprocess import call
from string import Template
from tempfile import NamedTemporaryFile
from xml.etree import ElementTree

from urllib.parse import urlparse, parse_qs
from urllib.parse import ParseResult
from urllib.parse import quote

import requests
import dbus
from wormhole._wordlist import PGPWordList
from _dbus_bindings import BUS_DAEMON_NAME, BUS_DAEMON_PATH, BUS_DAEMON_IFACE
import gi

gi.require_version('Gtk', '4.0')
#gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib

from .errors import NoBluezDbus, UnpoweredAdapter, NoAdapter
from .gpgmeh import fingerprint_from_keydata
from .gpgmeh import sign_keydata_and_encrypt
from .i18n import _
from .bluetoothutil import get_local_bt_address

log = logging.getLogger(__name__)


def mac_generate(key, data):
    mac = hmac.new(key, data, hashlib.sha256).hexdigest().upper()
    log.info("MAC of %r is %r", data[:20], mac[:20])
    # Arbitrary truncation to avoid a QR code size increase
    return mac[:20]


def mac_verify(key, data, mac):
    computed_mac = mac_generate(key, data)
    result = hmac.compare_digest(mac.upper(), computed_mac.upper())
    log.info("MAC of %r seems to be %r. Expected %r (%r)",
             data[:20], computed_mac[:20], mac[:20], result)
    return result



def get_window_handle(window):
    if not window:
        return ""
    if isinstance(window, str):
        return window

    try:
        import gi
        gi.require_version('Gtk', '4.0')
        from gi.repository import Gtk

        # If it has a custom portal_handle attribute (e.g. exported asynchronously on Wayland)
        if hasattr(window, "portal_handle"):
            return window.portal_handle

        surface = window.get_surface()
        if not surface:
            return ""

        # Try X11
        try:
            from gi.repository import GdkX11
            if isinstance(surface, GdkX11.X11Toplevel):
                return f"x11:{surface.get_xid()}"
        except Exception:
            pass

        # Try Wayland (if it wasn't exported yet or we want to try synchronously)
        try:
            from gi.repository import GdkWayland
            if isinstance(surface, GdkWayland.WaylandToplevel):
                if not hasattr(window, "_portal_handle_export_started"):
                    window._portal_handle_export_started = True
                    def on_handle_exported(toplevel, handle, *args):
                        window.portal_handle = f"wayland:{handle}"
                    surface.export_handle(on_handle_exported)
        except Exception:
            pass
    except Exception as e:
        log = logging.getLogger(__name__)
        log.debug("Could not get window handle: %s", e)

    return ""


def _email_portal(to, subject=None, body=None, files=None, parent_window=None):
    # The following checks are to ensure Python 2 compatibility
    if not hasattr(os, 'O_PATH'):
        os.O_PATH = 2097152
    if not hasattr(os, 'O_CLOEXEC'):
        os.O_CLOEXEC = 524288
    name = "org.freedesktop.portal.Desktop"
    path = "/org/freedesktop/portal/desktop"
    bus = dbus.SessionBus()
    try:
        proxy = bus.get_object(name, path)
    except dbus.exceptions.DBusException:
        log.debug("Desktop portal is not available")
        return None
    iface = "org.freedesktop.portal.Email"
    email = dbus.Interface(proxy, iface)
    parent_window_str = get_window_handle(parent_window)
    attrs = []
    # Even if we don't close the file descriptor it should not be a problem because
    # eventually at runtime it will be automatically closed.
    # Designing this class we took the recipes one as reference
    # https://gitlab.gnome.org/GNOME/recipes/blob/4afc9df6/src/gr-mail.c#L293
    if files:
        for file in files:
            fd = os.open(file, os.O_PATH | os.O_CLOEXEC)
            attrs.append(dbus.types.UnixFd(fd))
    opts = {"subject": subject, "address": to, "body": body, "attachment_fds": attrs}
    try:
        ret = email.ComposeEmail(parent_window_str, opts)
        return ret
    except TypeError:
        log.debug("Email portal is not available")
        return None


def _email_mailto(to, subject=None, body=None, files=None):
    url = "mailto:"
    url += "\"{0}\"".format(to)
    # Apparently we don't need to use urllib.parse.quote_plus
    if subject:
        url += "?subject={0}".format(subject)
    if body:
        if "?" in url:
            url += "&body={0}".format(quote(body))
        else:
            url += "?body={0}".format(quote(body))
    for file in files:
        if "?" in url:
            url += "&attach={0}".format(file)
        else:
            url += "?attach={0}".format(file)
    try:
        Gtk.show_uri(None, url, None)
        return True
    except GLib.GError as e:
        log.debug("mailto URI is probably not available: %s", e.message)
        return None


def _email_file(to, from_=None, subject=None,
                body=None,
                ccs=None, bccs=None,
                files=None, utf8=True):
    """Calls xdg-email with the appropriate options"""
    cmd = ['xdg-email']
    if utf8:
        cmd += ['--utf8']
    if subject:
        cmd += ['--subject', subject]
    if body:
        cmd += ['--body', body]
    for cc in ccs or []:
        cmd += ['--cc', cc]
    for bcc in bccs or []:
        cmd += ['--bcc', bcc]
    for file_ in files or []:
        cmd += ['--attach', file_]

    if not to:
        log.error("email_file: We are seeing an empty 'to': %r", to)

    cmd += [to]

    log.info("Running %s", cmd)
    retval = call(cmd)
    return retval


def _using_flatpak():
    """Check if we are inside flatpak"""
    return os.path.exists("/.flatpak-info")


def _fix_path_flatpak(files):
    """In Flatpak the only special path visible also from outside is /var/tmp/
    To be able to use the files from the host we change the path to the absolute one.
    This fix in the future may not be necessary because the portals should be able
    to automatically handle it."""
    part_1 = os.path.expanduser("~/.var/app/")
    app_id = "org.gnome.Keysign"
    part_2 = "cache/tmp/"
    flatpak_path = os.path.join(part_1, app_id, part_2)
    fixed_files = []
    if files:
        for file in files:
            base = os.path.basename(file)
            new_path = os.path.join(flatpak_path, base)
            shutil.move(file, new_path)
            fixed_files.append(new_path)
    return fixed_files


def send_email(to, subject=None, body=None, files=None, parent_window=None):
    """Tries to send the email using firstly the portal, then the xdg-email
    and as a last attempt the mailto uri"""
    if _using_flatpak():
        files = _fix_path_flatpak(files)

    if _email_portal(to, subject, body, files, parent_window=parent_window):
        return

    try:
        _email_file(to=to, subject=subject, body=body, files=files)
        return
    except FileNotFoundError:
        log.debug("xdg-email is not available")

    if _email_mailto(to, subject, body, files):
        return

    log.error("An error occurred trying to compose the email")


SUBJECT = 'Your signed key $fingerprint'
BODY = '''Hi $uid,


I have just signed your key

      $fingerprint


The certification is in an attachment encrypted to your key.
You can either drag the attachment into GNOME Keysign or
decrypt and import the attachment manually, e.g.

    gpg --decrypt  |  gpg --import


Thanks for letting me sign your key!
And, if you can, send me copy of your key with the new certifications.

--
GNOME Keysign
'''
DIVIDER = '\n--------\nTranslated version below\n--------\n'
BODY_TRANSLATED = _('''Hi $uid,


I have just signed your key

      $fingerprint


Thanks for letting me sign your key!

--
GNOME Keysign
''')


def sign_keydata_and_send(keydata, error_cb=None, parent_window=None, send_all_uids=False):
    """Creates, encrypts, and send signatures for each UID on the key
    
    You are supposed to give OpenPGP data which will be passed
    onto sign_keydata_and_encrypt.
    
    For the resulting signatures, emails are created and
    sent via send_email.
    
    Return value:  A tuple of a
    NamedTemporaryFiles used for saving the signatures and the plaintext which got encrypted.
    If you let the TemporaryFiles go out of scope they should get deleted.
    But don't delete too early as the MUA needs to pick them up.
    """
    log = logging.getLogger(__name__ + ':sign_keydata')

    fingerprint = fingerprint_from_keydata(keydata)
    # FIXME: We should rather use whatever GnuPG tells us
    keyid = fingerprint[-8:]
    # We list() the signatures, because we believe that it's more
    # acceptable if all key operations are done before we go ahead
    # and spawn an email client.
    log.info("About to create signatures for key with fpr %r", fingerprint)
    try:
        keydata = keydata.encode()
    except AttributeError:
        log.debug("keydata is probably already a bytes type")

    signed_uids = list(sign_keydata_and_encrypt(keydata, error_cb))
    emailable_uids = [(uid, enc, pt) for uid, enc, pt in signed_uids if uid.email and uid.email != 'unknown']
    emailless_uids = [(uid, enc, pt) for uid, enc, pt in signed_uids if not uid.email or uid.email == 'unknown']

    emailless_files = []
    if send_all_uids:
        for uid, encrypted_key, plaintext in emailless_uids:
            log.info("Using UID without email: %r", uid)
            tmpfile = NamedTemporaryFile(prefix='gnome-keysign-',
                                         suffix='.asc',
                                         delete=True)
            filename = tmpfile.name
            log.info('Writing keydata to %s', filename)
            tmpfile.write(encrypted_key)
            tmpfile.flush()
            tmpfile.file.close()
            emailless_files.append((uid, tmpfile, plaintext))

    emailable_files = []
    # If not send_all_uids, we only process emailable_uids as before (plus those without email if we were to process them as before? No, before it processed all and failed on empty email. We now filter them properly or just keep previous behavior if send_all_uids=False? Wait, if send_all_uids is False, we just do the old behavior. 
    # Actually, if we just process ALL UIDs like before when send_all_uids=False, it will try to send email and fail. That's what the user said: "or just do email as before".
    
    uids_to_email = emailable_uids if send_all_uids else signed_uids

    for uid, encrypted_key, plaintext in uids_to_email:
        log.info("Using UID: %r", uid)
        # We expect uid.uid to be a consumable string
        uid_str = uid.uid
        ctx = {
            'uid' : uid_str,
            'fingerprint': fingerprint,
            'keyid': keyid,
        }
        tmpfile = NamedTemporaryFile(prefix='gnome-keysign-',
                                     suffix='.asc',
                                     delete=True)
        filename = tmpfile.name
        log.info('Writing keydata to %s', filename)
        tmpfile.write(encrypted_key)
        # Interesting, sometimes it would not write the
        # whole thing out, so we better flush here
        tmpfile.flush()
        # If we close the actual file descriptor to free
        # resources. Calling tmpfile.close would get the file deleted.
        tmpfile.file.close()

        body = BODY
        # If the primary language is not English we append the translation
        if BODY != BODY_TRANSLATED:
            body += DIVIDER + BODY_TRANSLATED

        subject = Template(SUBJECT).safe_substitute(ctx)
        body = Template(body).safe_substitute(ctx)
        
        if send_all_uids:
            attachments = [filename] + [f[1].name for f in emailless_files]
        else:
            attachments = [filename]
        
        email_to = uid.email if uid.email != 'unknown' else ''
        send_email(email_to, subject, body, attachments, parent_window=parent_window)
        emailable_files.append((uid, tmpfile, plaintext))

    if send_all_uids and not emailable_uids and emailless_uids:
        log.warning("No emailable UIDs found. Certifications for email-less UIDs have been created but no emails were sent.")

    if send_all_uids:
        for uid, tmpfile, plaintext in emailable_files + emailless_files:
            yield tmpfile, plaintext
    else:
        for uid, tmpfile, plaintext in emailable_files:
            yield tmpfile, plaintext


def format_fingerprint(fpr):
    """Formats a given fingerprint (160bit, so 20 characters) in the
    GnuPG typical way
    """
    s = ''
    for i in range(10):
        # output 4 chars
        s += ''.join(fpr[4*i:4*i+4])
        # add extra space between the block
        if i == 4: s += '\n'
        # except at the end
        elif i < 9: s += ' '
    return s




def parse_barcode(barcode_string):
    """Parses information contained in a barcode

    It returns a dict with the parsed attributes.
    We expect the dict to contain at least a 'fingerprint'
    entry. Others might be added in the future.
    """
    # The string, currently, is of the form
    # openpgp4fpr:foobar?baz=qux#frag=val
    # Which urlparse handles perfectly fine.
    p = urlparse(barcode_string)
    log.debug("Parsed %r into %r", barcode_string, p)
    fpr = p.path
    query = parse_qs(p.query)
    fragments = parse_qs(p.fragment)
    rest = {}
    rest.update(query)
    rest.update(fragments)
    # We should probably ensure that we have only one
    # item for each parameter and flatten them accordingly.
    rest['fingerprint'] = fpr

    log.debug('Parsed barcode into %r', rest)
    return rest



FPR_PREFIX = "OPENPGP4FPR:"

def strip_fingerprint(input_string):
    '''Strips a fingerprint of any whitespaces and returns
    a clean version. It also drops the "OPENPGP4FPR:" prefix
    from the scanned QR-encoded fingerprints'''
    # The split removes the whitespaces in the string
    cleaned = ''.join(input_string.split())

    if cleaned.upper().startswith(FPR_PREFIX.upper()):
        cleaned = cleaned[len(FPR_PREFIX):]

    log.warning('Cleaned fingerprint to %s', cleaned)
    return cleaned




def download_key_http(address, port):
    netloc = "[%s]:%d" if ':' in address else "%s:%d"
    url = ParseResult(
        scheme='http',
        # This seems to work well enough with both IPv6 and IPv4
        netloc=netloc % (address, port),
        path='/',
        params='',
        query='',
        fragment='')
    u = url.geturl()
    log.debug("Starting HTTP request for %s", u)
    data = requests.get(u, timeout=5).content
    log.debug("finished downloading %d bytes", len(data))
    return data


def encode_message(message):
    """Serialize a string to json object and encode it in utf-8"""
    return json.dumps(message).encode("utf-8")


def decode_message(message):
    """deserialize a json returning a string"""
    return json.loads(message.decode("utf-8"))


def is_code_complete(code, length=2):
    if code[:1].isdigit():
        wl = PGPWordList()
        gc = wl.get_completions
        words = code.split("-", 1)[-1]
        return words in gc(words, length)
    else:
        return False


def fix_infobar(infobar):
    # Work around https://bugzilla.gnome.org/show_bug.cgi?id=710888
    # Taken from here https://phabricator.freedesktop.org/D1103#34aa2703
    def make_sure_revealer_does_nothing(widget):
        if not isinstance(widget, Gtk.Revealer):
            return
        widget.set_transition_type(Gtk.RevealerTransitionType.NONE)

    child = infobar.get_first_child()
    while child:
        make_sure_revealer_does_nothing(child)
        c = child.get_first_child()
        while c:
            make_sure_revealer_does_nothing(c)
            c = c.get_next_sibling()
        child = child.get_next_sibling()


def get_attachments(filename):
    mbox = mailbox.mbox(filename)
    attachments = []
    for message in mbox:
        # Check if there are attachments
        if message.is_multipart():
            for attach in message.get_payload()[1:]:
                attachments.append(attach.get_payload(decode=True))
    return attachments



"""Camera Portal client for XDG Desktop Portal camera access."""

import logging
import os
import re

import dbus
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

log = logging.getLogger(__name__)

PORTAL_BUS_NAME = "org.freedesktop.portal.Desktop"
PORTAL_OBJ_PATH = "/org/freedesktop/portal/desktop"
CAMERA_IFACE = "org.freedesktop.portal.Camera"
REQUEST_IFACE = "org.freedesktop.portal.Request"
PROPERTIES_IFACE = "org.freedesktop.DBus.Properties"

_request_token_counter = 0


def _using_flatpak():
    """Check if we are inside a Flatpak sandbox."""
    return os.path.exists("/.flatpak-info")


def is_camera_portal_available():
    """Check if the Camera Portal interface exists and has a camera."""
    result = False
    try:
        bus = dbus.SessionBus()
        proxy = bus.get_object(PORTAL_BUS_NAME, PORTAL_OBJ_PATH)
        props = dbus.Interface(proxy, PROPERTIES_IFACE)
        is_present = props.Get(CAMERA_IFACE, "IsCameraPresent")
        log.info("Camera portal IsCameraPresent: %s", is_present)
        result = bool(is_present)
    except dbus.exceptions.DBusException as e:
        log.debug("Camera portal not available: %s", e)
    return result


def request_camera_access(callback):
    """Request camera access asynchronously.
    
    Calls callback(True, pipewire_fd) on success,
    or callback(False, None) on failure/denial.
    """
    global _request_token_counter

    try:
        bus = dbus.SessionBus()
        portal = bus.get_object(PORTAL_BUS_NAME, PORTAL_OBJ_PATH)

        _request_token_counter += 1
        token = 'keysign_cam_%d' % _request_token_counter
        sender_name = re.sub(r'\.', r'_', bus.get_unique_name()[1:])
        request_path = (
            '/org/freedesktop/portal/desktop/request/%s/%s'
            % (sender_name, token))

        def on_response(response, results):
            if response == 0:
                log.info("Camera access granted")
                try:
                    fd_object = portal.OpenPipeWireRemote(
                        dbus.Dictionary(signature="sv"),
                        dbus_interface=CAMERA_IFACE)
                    pipewire_fd = fd_object.take()
                    log.info("Got PipeWire fd: %d", pipewire_fd)
                    callback(True, pipewire_fd)
                except dbus.exceptions.DBusException as e:
                    log.error("OpenPipeWireRemote failed: %s", e)
                    callback(False, None)
            else:
                log.info("Camera access denied (response=%d)", response)
                callback(False, None)

        bus.add_signal_receiver(
            on_response,
            signal_name='Response',
            dbus_interface=REQUEST_IFACE,
            path=request_path)

        portal.AccessCamera(
            dbus.Dictionary({'handle_token': dbus.String(token)},
                            signature="sv"),
            dbus_interface=CAMERA_IFACE)

        log.info("AccessCamera sent (token=%s)", token)

    except dbus.exceptions.DBusException as e:
        log.error("Failed to request camera access: %s", e)
        callback(False, None)



#!/usr/bin/env python
#    Copyright 2014 Tobias Mueller <muelli@cryptobitch.de>
#    Copyright 2015 Benjamin Berg <benjamin@sipsolutions.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.
import logging

import gi
gi.require_version('Gtk', '4.0')
gi.require_version('Gdk', '4.0')
from gi.repository import Gdk, Gtk, GObject
import qrcode

## It seems python3-cairo does not implement Surface.create_for_data
## https://bugs.freedesktop.org/show_bug.cgi?id=99855
## Also, the gi version of cairo seems to be mostly useless,
## so either pycairo or cairocffi needs to exist.
## Rumour has it, though, that cairocffi cannot work well together
## with the CairoContext exposed in the do_draw method that we are trying to
## overwrite.  So we're back to square one and need to take pycairo.
import cairo


log = logging.getLogger(__name__)

class QRImage(Gtk.DrawingArea):
    """An Image encoding data as a QR Code.
    The image tries to scale as big as possible.
    """
    
    def __init__(self, data='Default String', handle_events=True,
                       background=0xff, *args, **kwargs):
        """The QRImage widget inherits from Gtk.Image,
        but it probably cannot be used as one, as there
        is an event handler for resizing events which will
        overwrite to currently loaded image.
        
        You made set data now, or later simply via the property.
        
        handle_events can be set to False if the fullscreen
        window should not be created on click.
        
        The background can be set to 0x00 (or 0xff) creating a
        black (or white) background onto which the code is rendered.
        """
        super(QRImage, self).__init__(*args, **kwargs)
        self.log = logging.getLogger(__name__)

        self.background = background
        # We invert the background
        self.foreground = 0xff ^ background

        # The data to be rendered
        self._surface = None
        self.data = data
        self.set_draw_func(self.draw_func, None)

        self.handle_events = handle_events
        if handle_events:
            gesture = Gtk.GestureClick()
            gesture.connect('released', self.on_gesture_released)
            self.add_controller(gesture)

    def draw_func(self, drawing_area, cr, width, height, user_data):
        self.do_draw(cr, widget_width=width, widget_height=height)

    def on_gesture_released(self, gesture, n_press, x, y):
        button = gesture.get_current_button()
        if button == 1:
            w = FullscreenQRImageWindow(data=self.data)
            root = self.get_root()
            if root:
                w.set_transient_for(root)


    def do_draw(self, cr, widget_width, widget_height):
        """This scales the QR Code up to the widget's
        size. You may define your own size, but you must
        be careful not to cause too many resizing events.
        When you request a too big size, it may loop to death
        trying to fit the image.
        """
        data = self.data
        if widget_width is None or widget_height is None:
            box = self.get_allocation()
            widget_width, widget_height = box.width, box.height
        size = min(widget_width, widget_height)

        qrcode = self.qrcode
        img_size = qrcode.get_width()

        cr.save()

        background = self.background
        foreground = self.foreground

        # This seems to set the background,
        # but I'm not sure...
        cr.set_source_rgb(background, background, background)
        #cr.fill()
        # And have it painted
        cr.paint()
        # Now, I think we set the colour of the turtle
        # paint whatever is coming next.
        cr.set_source_rgb(foreground, foreground, foreground)
        # All of the rest I do not really understand,
        # but it seems to work reasonably well, without
        # weird PIL to Pixbuf hacks.
        cr.translate(widget_width / 2, widget_height / 2)
        scale = max(1, size / img_size)
        cr.scale(scale, scale)
        cr.translate(-img_size / 2, -img_size / 2)

        pattern = cairo.SurfacePattern(qrcode)
        pattern.set_filter(cairo.FILTER_NEAREST)
        cr.mask(pattern)

        cr.restore()

    def create_qrcode(self, data):
        log.debug('Encoding %s', data)
        code = qrcode.QRCode()

        code.add_data(data)

        matrix = code.get_matrix()
        size = len(matrix)
        stride = (size + 3) // 4 * 4
        log.debug("stride: %r  size: %r", stride, size)
        data = bytearray(stride * size)

        background = self.background
        foreground = self.foreground

        for x in range(size):
            for y in range(size):
                # Here we seem to be defining what
                # is going to be put on the surface.
                # I don't know what the semantic is,
                # though. Is 0 black? Or no modification
                # of the underlying background?
                # Anyway, this give us a nice white
                # QR Code.  Note that we do [y][x],
                # otherwise the generated code is diagonally
                # mirrored.
                if matrix[y][x]:
                    data[x + y * stride] = background
                else:
                    data[x + y * stride] = foreground

        surface = cairo.ImageSurface.create_for_data(data, cairo.FORMAT_A8, size, size, stride)

        return surface

    @property
    def qrcode(self):
        if self._surface is not None:
            return self._surface

        self._surface = self.create_qrcode(self.data)
        return self._surface

    def set_data(self, data):
        # FIXME: Full screen window is not updated in here ...
        self._data = data
        self._surface = None

        size = self.qrcode.get_width()
        self.set_size_request(size, size)

        self.queue_draw()
        
        self.set_tooltip_text(data)

    def get_data(self):
        return self._data

    data = GObject.Property(getter=get_data, setter=set_data)


def fullscreen_at_monitor(window, n):
    """Fullscreens a given window on the n-th monitor

    This is because Gtk's fullscreen_on_monitor seems to
    be buggy.
    http://stackoverflow.com/a/39386341/2015768
    """
    if hasattr(Gdk, 'Screen') and hasattr(Gdk.Screen, 'get_default'):
        screen = Gdk.Screen.get_default()
        monitor_n_geo = screen.get_monitor_geometry(n)
        x = monitor_n_geo.x
        y = monitor_n_geo.y
        window.move(x,y)
        window.fullscreen()
    else:
        window.fullscreen()


class FullscreenQRImageWindow(Gtk.Window):
    '''Displays a QRImage in a fullscreen window
    
    The window is supposed to close itself when a button is
    clicked.'''

    def __init__(self, data, *args, **kwargs):
        '''The data will be passed to the QRImage'''
        self.log = logging.getLogger(__name__)
        super(FullscreenQRImageWindow, self).__init__(*args, **kwargs)

        self.fullscreen()
        
        self.qrimage = QRImage(data=data, handle_events=False)
        self.qrimage.set_has_tooltip(False)
        self.set_child(self.qrimage)
        
        gesture = Gtk.GestureClick()
        gesture.connect('released', self.on_fullscreen_gesture_released)
        self.add_controller(gesture)
        key_controller = Gtk.EventControllerKey()
        key_controller.connect('key-released', self.on_fullscreen_key_released)
        self.add_controller(key_controller)

        self.present()

    def on_fullscreen_gesture_released(self, gesture, n_press, x, y):
        button = gesture.get_current_button()
        if button == 1:
            self.unfullscreen()
            self.hide()
            self.close()

    def on_fullscreen_key_released(self, controller, keyval, keycode, state):
        keyname = Gdk.keyval_name(keyval).lower()
        if keyname == 'escape' or keyname == 'f' or keyname == 'q':
            self.unfullscreen()
            self.hide()
            self.close()


def main(data):
    w = Gtk.Window()
    w.connect("delete-event", Gtk.main_quit)
    w.set_default_size(100,100)
    qr = QRImage(data)

    global fullscreen
    fullscreen = False

    def on_released(widget, event):
        global fullscreen
 
        if event.button == 1:
            fullscreen = not fullscreen
            if fullscreen:
                w.fullscreen()
            else:
                w.unfullscreen()
        
    #qr.connect('button-release-event', on_released)
    #qr.add_events(Gdk.EventMask.BUTTON_RELEASE_MASK | Gdk.EventMask.BUTTON_PRESS_MASK)
    w.set_child(qr)
    w.present()
    app = Gtk.Application()
    app.connect('activate', lambda app: (w.set_application(app), w.present()))
    app.run(None)

if __name__ == '__main__':
    import sys
    logging.basicConfig(level=logging.DEBUG)
    try:
        data = sys.argv[1]
    except:
        raise ValueError("Not Enough Arguments passed as data for the QR code encoding")
    main(data)



#!/usr/bin/env python
#    Copyright 2018 Ludovico de Nittis <aasonykk+gnome@gmail.com>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import sys

import dbus

if sys.version_info[0] >= 3:
    unicode = str


class AvahiConstants:
    SERVER_RUNNING = 2
    SERVER_COLLISION = 3

    ENTRY_GROUP_ESTABLISHED = 2
    ENTRY_GROUP_COLLISION = 3
    ENTRY_GROUP_FAILURE = 4

    PROTO_UNSPEC = -1

    IF_UNSPEC = -1

    LOOKUP_RESULT_LOCAL = 8

    DBUS_NAME = "org.freedesktop.Avahi"
    DBUS_INTERFACE_SERVER = DBUS_NAME + ".Server"
    DBUS_PATH_SERVER = "/"
    DBUS_INTERFACE_ENTRY_GROUP = DBUS_NAME + ".EntryGroup"
    DBUS_INTERFACE_SERVICE_BROWSER = DBUS_NAME + ".ServiceBrowser"

    @staticmethod
    def string_to_byte_array(s):
        if isinstance(s, unicode):
            s = s.encode('utf-8')

        r = []

        for c in s:
            if isinstance(c, int):
                # Python 3: iterating over bytes yields ints
                r.append(dbus.Byte(c))
            else:
                # Python 2: iterating over str yields str
                r.append(dbus.Byte(ord(c)))

        return r

    @staticmethod
    def dict_to_txt_array(txt_dict):
        l = []

        for k, v in txt_dict.items():
            if isinstance(k, unicode):
                k = k.encode('utf-8')

            if isinstance(v, unicode):
                v = v.encode('utf-8')

            l.append(AvahiConstants.string_to_byte_array(b"%s=%s" % (k, v)))

        return l



import sys
import os
import gi

gi.require_version('Gtk', '4.0')
from gi.repository import Gtk

# Setup Python path to import keysign
parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, parent_dir)
sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))

from keysign.keyconfirm import PreSignWidget

class DummyUID:
    def __init__(self, uid):
        self.uid = uid

class DummyKey:
    def __init__(self):
        self.fingerprint = "1234567890ABCDEF1234567890ABCDEF12345678"
        self.uidslist = [DummyUID("Test User <test@example.com>")]

def on_activate(app):
    window = Gtk.ApplicationWindow(application=app)
    window.set_title("Key Pre Sign Error Test")
    window.set_default_size(600, 400)

    key = DummyKey()
    psw = PreSignWidget(key)
    
    window.set_child(psw)
    window.present()
    
    # Programmatically trigger the error info bar
    try:
        raise RuntimeError("This is a simulated GPG error during certification!")
    except Exception as e:
        psw.infobar_errors.show(e)
        print("Error triggered! Click the 'Details' button in the UI to see the dialog.")

if __name__ == "__main__":
    app = Gtk.Application()
    app.connect('activate', on_activate)
    app.run(sys.argv)



#!/usr/bin/env python
#    Copyright 2016 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re
import os
import signal
import sys
from textwrap import dedent
from urllib.parse import unquote

import gi
gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')
from gi.repository import Gtk, GLib, Adw, Gdk, GObject
gi.require_version('Gst', '1.0')
from gi.repository import Gst
if __name__ == "__main__":
    from twisted.internet import gireactor
    gireactor.install()
from twisted.internet import reactor, threads
from twisted.internet.defer import inlineCallbacks
from wormhole.errors import WrongPasswordError, LonelyError

if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')


from .avahidiscovery import AvahiKeysignDiscoveryWithMac
from .discover import Discover
from .errors import NoBluezDbus, UnpoweredAdapter, NoAdapter
from .gpgmeh import openpgpkey_from_data, local_sign_keydata, GPGRuntimeError
from .i18n import _
from .keyfprscan import KeyFprScanWidget
from .keyconfirm import PreSignWidget
from .util import sign_keydata_and_send, fix_infobar, get_local_bt_address

log = logging.getLogger(__name__)


def remove_whitespace(s):
    cleaned = re.sub('[\s+]', '', s)
    return cleaned


class ReceiveApp:
    def __init__(self, builder=None):
        self.psw = None
        self.discovery = None
        self.log = logging.getLogger(__name__)

        widget_name = "receive_stack"
        if not builder:
            ui_file = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "receive4.ui")
            builder = Gtk.Builder()
            builder.add_objects_from_file(ui_file,
                [widget_name, 'confirm-button-image'])

        self.accept_button = builder.get_object("confirm_sign_button")

        old_scanner = builder.get_object("scanner_widget")
        old_scanner_parent = old_scanner.get_parent()

        scanner = KeyFprScanWidget() #builder=builder)
        scanner.connect("changed", self.on_code_changed)
        scanner.connect("barcode", self.on_barcode)

        if old_scanner_parent:
            old_scanner_parent.remove(old_scanner)
            # Hm. If we don't have an old parent, we never get to see
            # the newly created scanner. Weird.
            old_scanner_parent.append(scanner)

        receive_stack = builder.get_object(widget_name)
        # It needs to be show()n so that it can be made visible
        scanner.show()
        # FIXME: Use "stack_scanner_child" or so as identification
        # for the stack's scanner child to make it visible when the
        # app starts
        # receive_stack.set_visible_child(old_scanner_parent)
        self.scanner = scanner
        self.stack = receive_stack

        drop_target = Gtk.DropTarget.new(GObject.TYPE_STRING, Gdk.DragAction.COPY)
        drop_target.connect("drop", self.on_drop_data_received)
        self.scanner.add_controller(drop_target)

        self.discovery = AvahiKeysignDiscoveryWithMac()
        ib = builder.get_object('infobar_discovery')
        fix_infobar(ib)
        self.discovery.connect('list-changed', self.on_list_changed, ib)

        self.discover = None
        self.rb = builder.get_object('box50')
        self.result_label = builder.get_object("error_download_label")
        self.cancel_button = builder.get_object("cancel_download_button")
        self.redo_button = builder.get_object("redo_download_button")
        self.redo_button.connect("clicked", self.on_redo_button_clicked)
        self.cancel_button.connect("clicked", self.on_cancel_button_clicked)
        # Clear the "downloading" label
        builder.get_object("label10").set_label("")

        self.bt_usable = False

        # We call this in async because it can take several seconds to complete and we don't want
        # to stall the UI boot. Also we don't care about having this information immediately.
        threads.deferToThread(self.check_bt_availability)

    def on_drop_data_received(self, target, value, x, y):
        log.info("recv: Drag data rcvd: %s", value)
        dragged_data = unquote(value)
        if dragged_data.startswith("file://"):
            filename = dragged_data[7:].strip('\r\n\x00')  # remove file://, \r\n and NULL
            keydata = open(filename, 'br').read()

        elif dragged_data.startswith("-----BEGIN PGP PUBLIC KEY BLOCK-----"):
            # We assume a raw (well, armored) key to be passed
            keydata = dragged_data

        else:
            log.warning("We got a drag with neither file:// nor ----: %s", value)
            keydata = dragged_data

        self.on_keydata_downloaded(keydata)
        return True

    def on_redo_button_clicked(self, button):
        log.info("redo pressed")
        self.stack.remove(self.rb)
        self.discover.start()

    def on_cancel_button_clicked(self, button):
        log.info("cancel pressed")
        self.stack.remove(self.rb)

    def check_bt_availability(self):
        try:
            if get_local_bt_address():
                self.bt_usable = True
                log.debug("A working Bluetooth seems to be available")
            else:
                self.bt_usable = False
                log.debug("There is no usable Bluetooth")
        except NoBluezDbus as e:
            log.debug("Bluetooth service seems to be unavailable: %s", e)
        except NoAdapter as e:
            log.debug("Bluetooth adapter is not available: %s", e)
        except UnpoweredAdapter as e:
            log.debug("Bluetooth adapter is turned off: %s", e)

    def get_toplevel(self):
        if self.psw:
            return self.psw.get_root()
        return self.stack.get_root()
    def on_keydata_downloaded(self, keydata, pixbuf=None):
        log.debug("Downloaded keydata of length %d: %s", len(keydata), keydata[:50])
        key = openpgpkey_from_data(keydata)
        psw = PreSignWidget(key, pixbuf)
        psw.connect('sign-key-confirmed',
            self.on_sign_key_confirmed, keydata)
        self.stack.add_titled(psw, "presign", _("Sign Key"))
        psw.set_name("presign")
        psw.show()
        self.psw = psw
        self.stack.set_visible_child(self.psw)

    def on_message_received(self, key_data, success=True, message=None):
        if success:
            self.log.debug("message received")
            try:
                self.on_keydata_downloaded(key_data)
            except ValueError as ve:
                log.error(ve.args[0])
        else:
            self.stack.add_child(self.rb)
            self.result_label.set_label(dedent(message.__doc__))
            self.stack.set_visible_child(self.rb)

    def on_code_changed(self, scanner, entry):
        self.log.debug("Entry changed %r: %r", scanner, entry)
        text = entry.get_text()
        self._receive(text)

    def on_barcode(self, scanner, barcode, gstmessage, pixbuf):
        self.log.debug("Scanned barcode %r", barcode)
        self._receive(barcode)

    @inlineCallbacks
    def _receive(self, code):
        if self.discover:
            self.discover.stop()
        self.discover = Discover(code, self.discovery)
        msg_tuple = yield self.discover.start()
        key_data, success, message = msg_tuple
        if message == WrongPasswordError or message == LonelyError:
            # If a wrong password has been provided or we closed the connection
            # before a transfer. We do not display that to the user
            log.info("Waiting for another code")
            pass
        else:
            self.on_message_received(key_data, success, message)

    def on_sign_key_confirmed(self, keyPreSignWidget, key, keydata):
        self.log.debug ("Sign key confirmed! %r", key)
        # We need to prevent tmpfiles from going out of
        # scope too early so that they don't get deleted
        try:
            tmpfiles_plaintext = list(sign_keydata_and_send(keydata, parent_window=self.get_toplevel()))
        except GPGRuntimeError as e:
            self.log.exception("Something went wrong with signing the key")
            keyPreSignWidget.infobar_success.hide()
            keyPreSignWidget.infobar_errors.show(e)
        else:
            self.log.debug("sign keydata result: %r", tmpfiles_plaintext)
            # This is unzipping the list of tuples, e.g. [(1,2), (3,4)] becomes [(1,3), (2,4)]
            self.tmpfiles, plaintexts = zip(*tmpfiles_plaintext)

            keyPreSignWidget.infobar_errors.hide()
            keyPreSignWidget.infobar_success.show()

            def import_clicked(button):
                self.log.info("Import clicked")
                local_sign_keydata(keydata)
            keyPreSignWidget.infobar_import_button.connect("clicked", import_clicked)

            def save_as_clicked(button):
                self.log.info("Save as clicked")
                dialog = Gtk.FileChooserDialog(_("Select file for saving"),
                    self.get_toplevel(),
                    Gtk.FileChooserAction.SAVE,
                    (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                     Gtk.STOCK_SAVE, Gtk.ResponseType.OK)
                )
                response = dialog.run()
                if response == Gtk.ResponseType.OK:
                    filename = dialog.get_filename()
                    self.log.info("Saving file to: %r", filename)
                    with open(filename, 'wb') as f:
                        for p in plaintexts:
                            f.write(p)
                        for sigfile in self.tmpfiles:
                            pass
                            # Hrm. Those are the encrypted files, I think.
                            # We probably want to offer the plaintext versions, though
                            #f.write(open(sigfile, 'r').read())
                else:
                    self.log.info("Not saving file: %r", response)
                dialog.destroy()

            keyPreSignWidget.infobar_save_as_button.connect("clicked", save_as_clicked)

            # After the user has signed, we switch back to the scanner,
            # because currently, there is not much to do on the
            # key confirmation page.
            log.debug ("Signed the key: %r", self.tmpfiles)
            # self.stack.set_visible_child_name("scanner")

    def on_list_changed(self, discovery, number, userdata):
        """We show an infobar if we can only receive with Avahi and
        there are zero nearby servers"""
        ib = userdata
        if number == 0 and not self.bt_usable:
            ib.show()
        elif ib.is_visible():
            ib.hide()


class App(Adw.Application):
    def __init__(self, *args, **kwargs):
        super(App, self).__init__(*args, **kwargs)
        self.connect('activate', self.on_activate)
        self.log = logging.getLogger(__name__)

    def on_activate(self, app):
        ui_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "receive4.ui")
        builder = Gtk.Builder.new_from_file(ui_file)

        window = Adw.ApplicationWindow(application=app)
        window.connect("close-request", self.on_delete_window)
        window.set_title(_("Receive"))
        # window.set_size_request(600, 400)
        #window = self.builder.get_object("appwindow")
        
        self.receive = ReceiveApp(builder)
        receive_stack = self.receive.stack

        window.set_child(receive_stack)

        def on_realize(win):
            surface = win.get_surface()
            try:
                from gi.repository import GdkWayland
                if isinstance(surface, GdkWayland.WaylandToplevel):
                    def on_handle_exported(toplevel, handle, *args):
                        win.portal_handle = f"wayland:{handle}"
                    surface.export_handle(on_handle_exported)
            except Exception:
                pass
        window.connect("realize", on_realize)

        window.present()
        self.add_window(window)

    @staticmethod
    def on_delete_window(*args):
        reactor.callFromThread(reactor.stop)


def main(args=[]):
    log = logging.getLogger(__name__)
    log.debug('Running main with args: %s', args)
    if not args:
        args = []
    Gst.init(None)

    app = App(application_id="org.gnome.Keysign.Receive")
    try:
        GLib.unix_signal_add_full(GLib.PRIORITY_HIGH, signal.SIGINT,
                                  lambda *args: reactor.callFromThread(reactor.stop), None)
    except AttributeError:
        pass
    reactor.registerGApplication(app)
    reactor.run()

if __name__ == '__main__':
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
            format='%(name)s (%(levelname)s): %(message)s')
    sys.exit(main(sys.argv[1:]))
